"""
Multipurpose basic SIDD elements
"""

__classification__ = "UNCLASSIFIED"
__author__ = "Thomas McCullough"

from collections import OrderedDict
from typing import Union

import numpy

from sarpy.io.xml.base import Serializable, Arrayable, get_node_value, create_text_node, create_new_node, \
    find_children, find_first_child
from sarpy.io.xml.descriptors import SerializableDescriptor, IntegerDescriptor, \
    FloatDescriptor, FloatModularDescriptor, StringDescriptor, StringEnumDescriptor

from .base import DEFAULT_STRICT, FLOAT_FORMAT
from sarpy.io.complex.sicd_elements.blocks import XYZType as XYZTypeBase, XYZPolyType as XYZPolyTypeBase, \
    LatLonType as LatLonTypeBase, LatLonCornerType as LatLonCornerTypeBase, \
    RowColType as RowColIntTypeBase, RowColArrayElement as RowColArrayElementBase, \
    Poly1DType as Poly1DTypeBase, Poly2DType as Poly2DTypeBase, \
    LatLonCornerStringType as LatLonCornerStringTypeBase, LatLonArrayElementType as LatLonArrayElementTypeBase
from sarpy.io.complex.sicd_elements.ErrorStatistics import ErrorStatisticsType as ErrorStatisticsTypeBase
from sarpy.io.complex.sicd_elements.Radiometric import RadiometricType as RadiometricTypeBase
from sarpy.io.complex.sicd_elements.MatchInfo import MatchInfoType as MatchInfoTypeBase
from sarpy.io.complex.sicd_elements.GeoData import GeoInfoType as GeoInfoTypeBase
from sarpy.io.complex.sicd_elements.CollectionInfo import RadarModeType as RadarModeTypeBase

_len2_array_text = 'Expected array to be of length 2,\n\t' \
                   'and received `{}`'
_array_type_text = 'Expected array to be numpy.ndarray, list, or tuple,\n\t' \
                   'got `{}`'


############
# the SICommon namespace elements

class XYZType(XYZTypeBase):
    _child_xml_ns_key = {'X': 'sicommon', 'Y': 'sicommon', 'Z': 'sicommon'}


class LatLonType(LatLonTypeBase):
    _child_xml_ns_key = {'Lat': 'sicommon', 'Lon': 'sicommon'}


class LatLonCornerType(LatLonCornerTypeBase):
    _child_xml_ns_key = {'Lat': 'sicommon', 'Lon': 'sicommon'}


class LatLonCornerStringType(LatLonCornerStringTypeBase):
    _child_xml_ns_key = {'Lat': 'sicommon', 'Lon': 'sicommon'}


class LatLonArrayElementType(LatLonArrayElementTypeBase):
    _child_xml_ns_key = {'Lat': 'sicommon', 'Lon': 'sicommon'}


class RangeAzimuthType(Serializable, Arrayable):
    """
    Represents range and azimuth.
    """
    _fields = ('Range', 'Azimuth')
    _required = ('Range', 'Azimuth')
    _numeric_format = {key: FLOAT_FORMAT for key in _fields}
    _child_xml_ns_key = {'Range': 'sicommon', 'Azimuth': 'sicommon'}
    # Descriptor
    Range = FloatDescriptor(
        'Range', _required, strict=DEFAULT_STRICT,
        docstring='The range in meters.')  # type: float
    Azimuth = FloatDescriptor(
        'Azimuth', _required, strict=DEFAULT_STRICT,
        docstring='The azimuth in degrees.')  # type: float

    def __init__(self, Range=None, Azimuth=None, **kwargs):
        """

        Parameters
        ----------
        Range : float
        Azimuth : float
        kwargs
        """

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.Range = Range
        self.Azimuth = Azimuth
        super(RangeAzimuthType, self).__init__(**kwargs)

    def get_array(self, dtype=numpy.float64):
        """
        Gets an array representation of the class instance.

        Parameters
        ----------
        dtype : str|numpy.dtype|numpy.number
            numpy data type of the return

        Returns
        -------
        numpy.ndarray
            array of the form [Range, Azimuth]
        """

        return numpy.array([self.Range, self.Azimuth], dtype=dtype)

    @classmethod
    def from_array(cls, array):
        """
        Create from an array type entry.

        Parameters
        ----------
        array: numpy.ndarray|list|tuple
            assumed [Range, Azimuth]

        Returns
        -------
        RangeAzimuthType
        """
        if array is None:
            return None
        if isinstance(array, (numpy.ndarray, list, tuple)):
            if len(array) < 2:
                raise ValueError(_len2_array_text.format(array))
            return cls(Range=array[0], Azimuth=array[1])
        raise ValueError(_array_type_text.format(type(array)))


class AngleMagnitudeType(Serializable, Arrayable):
    """
    Represents a magnitude and angle.
    """

    _fields = ('Angle', 'Magnitude')
    _required = ('Angle', 'Magnitude')
    _numeric_format = {key: FLOAT_FORMAT for key in _fields}
    _child_xml_ns_key = {'Angle': 'sicommon', 'Magnitude': 'sicommon'}
    # Descriptor
    Angle = FloatModularDescriptor(
        'Angle', 180.0, _required, strict=DEFAULT_STRICT,
        docstring='The angle.')  # type: float
    Magnitude = FloatDescriptor(
        'Magnitude', _required, strict=DEFAULT_STRICT, bounds=(0.0, None),
        docstring='The magnitude.')  # type: float

    def __init__(self, Angle=None, Magnitude=None, **kwargs):
        """

        Parameters
        ----------
        Angle : float
        Magnitude : float
        kwargs
        """

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.Angle = Angle
        self.Magnitude = Magnitude
        super(AngleMagnitudeType, self).__init__(**kwargs)

    def get_array(self, dtype=numpy.float64):
        """
        Gets an array representation of the class instance.

        Parameters
        ----------
        dtype : str|numpy.dtype|numpy.number
            numpy data type of the return

        Returns
        -------
        numpy.ndarray
            array of the form [Angle, Magnitude]
        """

        return numpy.array([self.Angle, self.Magnitude], dtype=dtype)

    @classmethod
    def from_array(cls, array):
        """
        Create from an array type entry.

        Parameters
        ----------
        array: numpy.ndarray|list|tuple
            assumed [Angle, Magnitude]

        Returns
        -------
        AngleMagnitudeType
        """

        if array is None:
            return None
        if isinstance(array, (numpy.ndarray, list, tuple)):
            if len(array) < 2:
                raise ValueError(_len2_array_text.format(array))
            return cls(Angle=array[0], Magnitude=array[1])
        raise ValueError(_array_type_text.format(type(array)))


class RowColIntType(RowColIntTypeBase):
    _child_xml_ns_key = {'Row': 'sicommon', 'Col': 'sicommon'}


class RowColArrayElement(RowColArrayElementBase):
    _child_xml_ns_key = {'Row': 'sicommon', 'Col': 'sicommon'}


class RowColDoubleType(Serializable, Arrayable):
    _fields = ('Row', 'Col')
    _required = _fields
    _numeric_format = {key: FLOAT_FORMAT for key in _fields}
    _child_xml_ns_key = {'Row': 'sicommon', 'Col': 'sicommon'}
    # Descriptors
    Row = FloatDescriptor(
        'Row', _required, strict=True, docstring='The Row attribute.')  # type: float
    Col = FloatDescriptor(
        'Col', _required, strict=True, docstring='The Column attribute.')  # type: float

    def __init__(self, Row=None, Col=None, **kwargs):
        """
        Parameters
        ----------
        Row : float
        Col : float
        kwargs
        """

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.Row, self.Col = Row, Col
        super(RowColDoubleType, self).__init__(**kwargs)

    def get_array(self, dtype=numpy.float64):
        """
        Gets an array representation of the class instance.

        Parameters
        ----------
        dtype : str|numpy.dtype|numpy.number
            numpy data type of the return

        Returns
        -------
        numpy.ndarray
            array of the form [Row, Col]
        """

        return numpy.array([self.Row, self.Col], dtype=dtype)

    @classmethod
    def from_array(cls, array):
        """
        Create from an array type entry.

        Parameters
        ----------
        array: numpy.ndarray|list|tuple
            assumed [Row, Col]

        Returns
        -------
        RowColDoubleType
        """
        if array is None:
            return None
        if isinstance(array, (numpy.ndarray, list, tuple)):
            if len(array) < 2:
                raise ValueError(_len2_array_text.format(array))
            return cls(Row=array[0], Col=array[1])
        raise ValueError(_array_type_text.format(type(array)))


class Poly1DType(Poly1DTypeBase):
    _child_xml_ns_key = {'Coefs': 'sicommon'}


class Poly2DType(Poly2DTypeBase):
    _child_xml_ns_key = {'Coefs': 'sicommon'}


class XYZPolyType(XYZPolyTypeBase):
    _child_xml_ns_key = {'X': 'sicommon', 'Y': 'sicommon', 'Z': 'sicommon'}


class ErrorStatisticsType(ErrorStatisticsTypeBase):
    _child_xml_ns_key = {'CompositeSCP': 'sicommon', 'Components': 'sicommon', 'AdditionalParms': 'sicommon'}


class RadiometricType(RadiometricTypeBase):
    _child_xml_ns_key = {
        'NoiseLevel': 'sicommon', 'RCSSFPoly': 'sicommon', 'SigmaZeroSFPoly': 'sicommon',
        'BetaZeroSFPoly': 'sicommon', 'GammaZeroSFPoly': 'sicommon'}


class MatchInfoType(MatchInfoTypeBase):
    _child_xml_ns_key = {'NumMatchTypes': 'sicommon', 'MatchTypes': 'sicommon'}


class GeoInfoType(GeoInfoTypeBase):
    _child_xml_ns_key = {
        'name': None, 'Descriptions': 'sicommon', 'Point': 'sicommon', 'Line': 'sicommon',
        'Polygon': 'sicommon'}


class RadarModeType(RadarModeTypeBase):
    """
    Radar mode type container class
    """

    _child_xml_ns_key = {'ModeType': 'sicommon', 'ModeID': 'sicommon'}
    # other class variable
    _MODE_TYPE_VALUES = RadarModeTypeBase._MODE_TYPE_VALUES + ('SCANSAR',)
    # descriptors
    ModeType = StringEnumDescriptor(
        'ModeType', _MODE_TYPE_VALUES, RadarModeTypeBase._required, strict=True,
        docstring="The Radar imaging mode.")  # type: str


class ReferencePointType(Serializable):
    """
    A reference point.
    """

    _fields = ('ECEF', 'Point', 'name')
    _required = ('ECEF', 'Point')
    _set_as_attribute = ('name', )
    _child_xml_ns_key = {'ECEF': 'sicommon', 'Point': 'sicommon'}
    # Descriptor
    ECEF = SerializableDescriptor(
        'ECEF', XYZType, _required, strict=DEFAULT_STRICT,
        docstring='The ECEF coordinates of the reference point.')  # type: XYZType
    Point = SerializableDescriptor(
        'Point', RowColDoubleType, _required, strict=DEFAULT_STRICT,
        docstring='The pixel coordinates of the reference point.')  # type: RowColDoubleType
    name = StringDescriptor(
        'name', _required, strict=DEFAULT_STRICT,
        docstring='Used for implementation specific signifier for the reference point.')  # type: Union[None, str]

    def __init__(self, ECEF=None, Point=None, name=None, **kwargs):
        """

        Parameters
        ----------
        ECEF : XYZType|numpy.ndarray|list|tuple
        Point : RowColDoubleType|numpy.ndarray|list|tuple
        name : None|str
        kwargs
        """

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.ECEF = ECEF
        self.Point = Point
        self.name = name
        super(ReferencePointType, self).__init__(**kwargs)

# The end of the SICommon namespace
#####################


#################
# Filter Type

class PredefinedFilterType(Serializable):
    """
    The predefined filter type.
    """
    _fields = ('DatabaseName', 'FilterFamily', 'FilterMember')
    _required = ()
    # Descriptor
    DatabaseName = StringEnumDescriptor(
        'DatabaseName', ('BILINEAR', 'CUBIC', 'LAGRANGE', 'NEAREST NEIGHBOR'),
        _required, strict=DEFAULT_STRICT,
        docstring='The filter name.')  # type: str
    FilterFamily = IntegerDescriptor(
        'FilterFamily', _required, strict=DEFAULT_STRICT,
        docstring='The filter family number.')  # type: int
    FilterMember = IntegerDescriptor(
        'FilterMember', _required, strict=DEFAULT_STRICT,
        docstring='The filter member number.')  # type: int

    def __init__(self, DatabaseName=None, FilterFamily=None, FilterMember=None, **kwargs):
        """

        Parameters
        ----------
        DatabaseName : None|str
        FilterFamily : None|int
        FilterMember : None|int
        kwargs
        """

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.DatabaseName = DatabaseName
        self.FilterFamily = FilterFamily
        self.FilterMember = FilterMember
        super(PredefinedFilterType, self).__init__(**kwargs)


class _CustomType(Serializable, Arrayable):
    """
    A custom filter array.
    """
    __slots__ = ('_coefs', )
    _fields = ('Coefs', )
    _required = ('Coefs', )
    _numeric_format = {'Coefs': FLOAT_FORMAT}
    _var0 = "x"
    _nvar0 = "numXs"
    _var1 = "y"
    _nvar1 = "numYs"

    def __init__(self, Coefs=None, **kwargs):
        """

        Parameters
        ----------
        Coefs : numpy.ndarray|list|tuple
        kwargs
        """
        self._coefs = None
        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.Coefs = Coefs
        super().__init__(**kwargs)

    def _shape0(self):
        return self._coefs.shape[0]

    def _shape1(self):
        return self._coefs.shape[1]

    @property
    def Coefs(self):
        """
        numpy.ndarray: The two-dimensional filter coefficient array of dtype=float64. Assignment object must be a
        two-dimensional numpy.ndarray, or naively convertible to one.

        .. Note:: this returns the direct coefficient array. Use the `get_array()` method to get a copy of the
            coefficient array of specified data type.
        """

        return self._coefs

    @Coefs.setter
    def Coefs(self, value):
        if value is None:
            raise ValueError('The coefficient array must be defined.')

        if isinstance(value, (list, tuple)):
            value = numpy.array(value, dtype=numpy.float64)

        if not isinstance(value, numpy.ndarray):
            raise ValueError(
                'Coefs must be a list or numpy.ndarray. Received type {}.'.format(type(value)))
        elif len(value.shape) != 2:
            raise ValueError(
                'Coefs must be two-dimensional. Received numpy.ndarray '
                'of shape {}.'.format(value.shape))
        elif not value.dtype.name == 'float64':
            value = numpy.asarray(value, dtype=numpy.float64)
        self._coefs = value

    def __getitem__(self, item):
        return self._coefs[item]

    def __setitem__(self, item, value):
        self._coefs[item] = value

    @classmethod
    def from_array(cls, array):
        if array is None:
            return None
        return cls(Coefs=array)

    def get_array(self, dtype=numpy.float64):
        """
        Gets **a copy** of the coefficent array of specified data type.

        Parameters
        ----------
        dtype : str|numpy.dtype|numpy.number
            numpy data type of the return

        Returns
        -------
        numpy.ndarray
            two-dimensional coefficient array
        """

        return numpy.array(self._coefs, dtype=dtype)

    @classmethod
    def from_node(cls, node, xml_ns, ns_key=None, kwargs=None):
        filter_coefs_node = find_first_child(node, "FilterCoefficients", xml_ns, ns_key)
        num_x = int(filter_coefs_node.attrib[cls._nvar0])
        num_y = int(filter_coefs_node.attrib[cls._nvar1])
        coefs = numpy.zeros((num_x, num_y), dtype=numpy.float64)
        ckey = cls._child_xml_ns_key.get('Coefs', ns_key)
        coef_nodes = find_children(filter_coefs_node, 'Coef', xml_ns, ckey)
        for cnode in coef_nodes:
            ind1 = int(cnode.attrib[cls._var0])
            ind2 = int(cnode.attrib[cls._var1])
            val = float(get_node_value(cnode))
            coefs[ind1, ind2] = val
        return cls(Coefs=coefs)

    def to_node(self, doc, tag, ns_key=None, parent=None, check_validity=False, strict=DEFAULT_STRICT, exclude=()):
        if parent is None:
            parent = doc.getroot()
        if ns_key is None:
            node = create_new_node(doc, tag, parent=parent)
        else:
            node = create_new_node(doc, '{}:{}'.format(ns_key, tag), parent=parent)
        fc_tag = "FilterCoefficients" if ns_key is None else f"{ns_key}:FilterCoefficients"
        filter_coefs_node = create_new_node(doc, fc_tag, parent=node)

        if 'Coefs' in self._child_xml_ns_key:
            ctag = '{}:Coef'.format(self._child_xml_ns_key['Coefs'])
        elif ns_key is not None:
            ctag = '{}:Coef'.format(ns_key)
        else:
            ctag = 'Coef'

        filter_coefs_node.attrib[self._nvar0] = str(self._shape0())
        filter_coefs_node.attrib[self._nvar1] = str(self._shape1())
        fmt_func = self._get_formatter('Coefs')
        for i, val1 in enumerate(self._coefs):
            for j, val in enumerate(val1):
                # if val != 0.0:  # should we serialize it sparsely?
                cnode = create_text_node(doc, ctag, fmt_func(val), parent=filter_coefs_node)
                cnode.attrib[self._var0] = str(i)
                cnode.attrib[self._var1] = str(j)
        return node

    def to_dict(self,  check_validity=False, strict=DEFAULT_STRICT, exclude=()):
        out = OrderedDict()
        out['Coefs'] = self.Coefs.tolist()
        return out


class KernelCustomType(_CustomType):
    """
    A custom filter kernel array.
    """
    _var0 = "row"
    _nvar0 = "numRows"
    _var1 = "col"
    _nvar1 = "numCols"

    @property
    def numRows(self):
        return self._shape0()

    @property
    def numCols(self):
        return self._shape1()


class FilterKernelType(Serializable):
    """
    The filter kernel parameters. Provides the specifics for **either** a predefined or custom
    filter kernel.
    """

    _fields = ('Predefined', 'Custom')
    _required = ()
    _choice = ({'required': True, 'collection': ('Predefined', 'Custom')}, )
    # Descriptor
    Predefined = SerializableDescriptor(
        'Predefined', PredefinedFilterType, _required, strict=DEFAULT_STRICT,
        docstring='')  # type: PredefinedFilterType
    Custom = SerializableDescriptor(
        'Custom', KernelCustomType, _required, strict=DEFAULT_STRICT,
        docstring='The custom filter kernel')  # type: KernelCustomType

    def __init__(self, Predefined=None, Custom=None, **kwargs):
        """

        Parameters
        ----------
        Predefined : None|PredefinedFilterType
        Custom : None|KernelCustomType
        kwargs
        """

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.Predefined = Predefined
        self.Custom = Custom
        super(FilterKernelType, self).__init__(**kwargs)


class BankCustomType(_CustomType):
    """
    A custom filter bank array.
    """
    _var0 = "phasing"
    _nvar0 = "numPhasings"
    _var1 = "point"
    _nvar1 = "numPoints"

    @property
    def numPhasings(self):
        return self._shape0()

    @property
    def numPoints(self):
        return self._shape1()


class FilterBankType(Serializable):
    """
    The filter bank type. Provides the specifics for **either** a predefined or custom filter bank.
    """

    _fields = ('Predefined', 'Custom')
    _required = ()
    _choice = ({'required': True, 'collection': ('Predefined', 'Custom')}, )
    # Descriptor
    Predefined = SerializableDescriptor(
        'Predefined', PredefinedFilterType, _required, strict=DEFAULT_STRICT,
        docstring='The predefined filter bank type.')  # type: PredefinedFilterType
    Custom = SerializableDescriptor(
        'Custom', BankCustomType, _required, strict=DEFAULT_STRICT,
        docstring='The custom filter bank.')  # type: BankCustomType

    def __init__(self, Predefined=None, Custom=None, **kwargs):
        """

        Parameters
        ----------
        Predefined : None|PredefinedFilterType
        Custom : None|BankCustomType
        kwargs
        """

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.Predefined = Predefined
        self.Custom = Custom
        super(FilterBankType, self).__init__(**kwargs)


class FilterType(Serializable):
    """
    Filter parameters for a variety of purposes. Provides **either** the filter bank or
    filter kernel parameters.
    """

    _fields = ('FilterName', 'FilterKernel', 'FilterBank', 'Operation')
    _required = ('FilterName', 'Operation')
    _choice = ({'required': True, 'collection': ('FilterKernel', 'FilterBank')}, )
    # Descriptor
    FilterName = StringDescriptor(
        'FilterName', _required, strict=DEFAULT_STRICT,
        docstring='The name of the filter.')  # type : str
    FilterKernel = SerializableDescriptor(
        'FilterKernel', FilterKernelType, _required, strict=DEFAULT_STRICT,
        docstring='The filter kernel.')  # type: FilterKernelType
    FilterBank = SerializableDescriptor(
        'FilterBank', FilterBankType, _required, strict=DEFAULT_STRICT,
        docstring='The filter bank.')  # type: FilterBankType
    Operation = StringEnumDescriptor(
        'Operation', ('CONVOLUTION', 'CORRELATION'), _required, strict=DEFAULT_STRICT,
        docstring='')  # type: str

    def __init__(self, FilterName=None, FilterKernel=None, FilterBank=None, Operation=None, **kwargs):
        """

        Parameters
        ----------
        FilterName : str
        FilterKernel : None|FilterKernelType
        FilterBank : None|FilterBankType
        Operation : str
        kwargs
        """

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.FilterName = FilterName
        self.FilterKernel = FilterKernel
        self.FilterBank = FilterBank
        self.Operation = Operation
        super(FilterType, self).__init__(**kwargs)


################
# NewLookupTableType


class PredefinedLookupType(Serializable):
    """
    The predefined lookup table type. Allows for reference **either** by name, or family/member id number.
    """
    _fields = ('DatabaseName', 'RemapFamily', 'RemapMember')
    _required = ()
    # Descriptor
    DatabaseName = StringDescriptor(
        'DatabaseName', _required, strict=DEFAULT_STRICT,
        docstring='Database name of LUT to use.')  # type: str
    RemapFamily = IntegerDescriptor(
        'RemapFamily', _required, strict=DEFAULT_STRICT,
        docstring='The lookup family number.')  # type: int
    RemapMember = IntegerDescriptor(
        'RemapMember', _required, strict=DEFAULT_STRICT,
        docstring='The lookup member number.')  # type: int

    def __init__(self, DatabaseName=None, RemapFamily=None, RemapMember=None, **kwargs):
        """

        Parameters
        ----------
        DatabaseName : None|str
        RemapFamily : None|int
        RemapMember : None|int
        kwargs
        """

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.DatabaseName = DatabaseName
        self.RemapFamily = RemapFamily
        self.RemapMember = RemapMember
        super(PredefinedLookupType, self).__init__(**kwargs)


class LUTInfoType(Serializable, Arrayable):
    """
    The lookup table - basically just a one or two-dimensional unsigned integer array of bit depth 8 or 16.
    """
    __slots__ = ('_lut_values', )
    _fields = ('LUTValues', 'numLuts', 'size')
    _required = ('LUTValues', 'numLuts', 'size')

    def __init__(self, LUTValues=None, **kwargs):
        """

        Parameters
        ----------
        LUTValues : numpy.ndarray
            The dtype must be `uint8` or `uint16`, and the dimension must be one or two.
        kwargs
        """

        self._lut_values = None
        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.LUTValues = LUTValues
        super(LUTInfoType, self).__init__(**kwargs)

    @property
    def LUTValues(self):
        """
        numpy.ndarray: the two-dimensional look-up table, where the dtype must be `uint8` or `uint16`.
        The first dimension should correspond to entries (i.e. size of the lookup table), and the
        second dimension should correspond to bands (i.e. number of bands).
        """

        return self._lut_values

    @LUTValues.setter
    def LUTValues(self, value):
        if value is None:
            self._lut_values = None
            return
        if isinstance(value, (tuple, list)):
            value = numpy.array(value, dtype=numpy.uint8)
        if not isinstance(value, numpy.ndarray) or value.dtype.name not in ('uint8', 'uint16'):
            raise ValueError(
                'LUTValues for class LUTInfoType must be a numpy.ndarray of dtype uint8 or uint16.')
        if value.ndim != 2:
            raise ValueError('LUTValues for class LUTInfoType must be two-dimensional.')
        self._lut_values = value

    @property
    def size(self):
        """
        int: the size of each lookup table
        """
        if self._lut_values is None:
            return 0
        else:
            return self._lut_values.shape[0]

    @property
    def numLUTs(self):
        """
        int: The number of lookup tables
        """
        if self._lut_values is None:
            return 0
        else:
            return self._lut_values.shape[1]

    def __len__(self):
        if self._lut_values is None:
            return 0
        return self._lut_values.shape[0]

    def __getitem__(self, item):
        return self._lut_values[item]

    @classmethod
    def from_array(cls, array):
        """
        Create from the lookup table array.

        Parameters
        ----------
        array: numpy.ndarray|list|tuple
            Must be two-dimensional. If not a numpy.ndarray, this will be naively
            interpreted as `uint8`.

        Returns
        -------
        LUTInfoType
        """
        if array is None:
            return None
        return cls(LUTValues=array)

    def get_array(self, dtype=numpy.uint8):
        """
        Gets **a copy** of the coefficent array of specified data type.

        Parameters
        ----------
        dtype : str|numpy.dtype|numpy.number
            numpy data type of the return

        Returns
        -------
        numpy.ndarray
            the lookup table array
        """

        return numpy.array(self._lut_values, dtype=dtype)

    @classmethod
    def from_node(cls, node, xml_ns, ns_key=None, kwargs=None):
        dim1 = int(node.attrib['size'])
        dim2 = int(node.attrib['numLuts'])
        arr = numpy.zeros((dim1, dim2), dtype=numpy.uint16)

        lut_key = cls._child_xml_ns_key.get('LUTValues', ns_key)
        lut_nodes = find_children(node, 'LUTValues', xml_ns, lut_key)
        for i, lut_node in enumerate(lut_nodes):
            arr[:, i] = [str(el) for el in get_node_value(lut_node)]
        if numpy.max(arr) < 256:
            arr = numpy.asarray(arr, dtype=numpy.uint8)
        return cls(LUTValues=arr)

    def to_node(self, doc, tag, ns_key=None, parent=None, check_validity=False, strict=DEFAULT_STRICT, exclude=()):
        def make_entry(arr):
            value = ' '.join(str(el) for el in arr)
            entry = create_text_node(doc, ltag, value, parent=node)
            entry.attrib['lut'] = str(arr.size)

        if parent is None:
            parent = doc.getroot()

        if ns_key is None:
            node = create_new_node(doc, tag, parent=parent)
        else:
            node = create_new_node(doc, '{}:{}'.format(ns_key, tag), parent=parent)

        if 'LUTValues' in self._child_xml_ns_key:
            ltag = '{}:LUTValues'.format(self._child_xml_ns_key['LUTValues'])
        elif ns_key is not None:
            ltag = '{}:LUTValues'.format(ns_key)
        else:
            ltag = 'LUTValues'

        if self._lut_values is not None:
            node.attrib['numLuts'] = str(self.numLUTs)
            node.attrib['size'] = str(self.size)
            if self._lut_values.ndim == 1:
                make_entry(self._lut_values)
            else:
                for j in range(self._lut_values.shape[1]):
                    make_entry(self._lut_values[:, j])
        return node

    def to_dict(self,  check_validity=False, strict=DEFAULT_STRICT, exclude=()):
        out = OrderedDict()
        if self.LUTValues is not None:
            out['LUTValues'] = self.LUTValues.tolist()
        return out


class CustomLookupType(Serializable):
    """
    A custom lookup table.
    """

    _fields = ('LUTInfo', )
    _required = ('LUTInfo', )
    # Descriptor
    LUTInfo = SerializableDescriptor(
        'LUTInfo', LUTInfoType, _required, strict=DEFAULT_STRICT,
        docstring='The lookup table.')  # type: LUTInfoType

    def __init__(self, LUTInfo=None, **kwargs):
        """

        Parameters
        ----------
        LUTInfo: LUTInfoType|numpy.ndarray|list|tuple
        kwargs
        """

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.LUTInfo = LUTInfo
        super(CustomLookupType, self).__init__(**kwargs)


class NewLookupTableType(Serializable):
    """
    The lookup table. Allows **either** a reference to a prefined lookup table, or
    custom lookup table array.
    """

    _fields = ('LUTName', 'Predefined', 'Custom')
    _required = ('LUTName', )
    _choice = ({'required': True, 'collection': ('Predefined', 'Custom')}, )
    # Descriptor
    LUTName = StringDescriptor(
        'LUTName', _required, strict=DEFAULT_STRICT,
        docstring='The lookup table name')  # type: str
    Predefined = SerializableDescriptor(
        'Predefined', PredefinedLookupType, _required, strict=DEFAULT_STRICT,
        docstring='')  # type: PredefinedLookupType
    Custom = SerializableDescriptor(
        'Custom', CustomLookupType, _required, strict=DEFAULT_STRICT,
        docstring='')  # type: CustomLookupType

    def __init__(self, LUTName=None, Predefined=None, Custom=None, **kwargs):
        """

        Parameters
        ----------
        LUTName : str
        Predefined : None|PredefinedLookupType
        Custom : None|CustomLookupType
        kwargs
        """

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.LUTName = LUTName
        self.Predefined = Predefined
        self.Custom = Custom
        super(NewLookupTableType, self).__init__(**kwargs)



"""
The SRP definition for CPHD 0.3.
"""

__classification__ = "UNCLASSIFIED"
__author__ = "Thomas McCullough"

from typing import Union
import numpy

from sarpy.io.phase_history.cphd1_elements.base import DEFAULT_STRICT
from sarpy.io.xml.base import Serializable
from sarpy.io.xml.descriptors import SerializableDescriptor, IntegerEnumDescriptor


class FxParametersType(Serializable):
    """
    The FX vector parameters.
    """

    _fields = ('Fx0', 'Fx_SS', 'Fx1', 'Fx2')
    _required = _fields
    # descriptors
    Fx0 = IntegerEnumDescriptor(
        'Fx0', (8, ), _required, strict=DEFAULT_STRICT, default_value=8,
        docstring='The size of the Fx0 field')  # type: int
    Fx_SS = IntegerEnumDescriptor(
        'Fx_SS', (8, ), _required, strict=DEFAULT_STRICT, default_value=8,
        docstring='The size of the Fx_SS field')  # type: int
    Fx1 = IntegerEnumDescriptor(
        'Fx1', (8, ), _required, strict=DEFAULT_STRICT, default_value=8,
        docstring='The size of the Fx1 field')  # type: int
    Fx2 = IntegerEnumDescriptor(
        'Fx2', (8, ), _required, strict=DEFAULT_STRICT, default_value=8,
        docstring='The size of the Fx2 field')  # type: int

    def __init__(self, Fx0=8, Fx_SS=8, Fx1=8, Fx2=8, **kwargs):
        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.Fx0 = Fx0
        self.Fx1 = Fx1
        self.Fx2 = Fx2
        self.Fx_SS = Fx_SS
        super(FxParametersType, self).__init__(**kwargs)

    @staticmethod
    def get_size():
        """
        The size in bytes of this component of the vector.

        Returns
        -------
        int
        """

        return 32

    def get_position_offset_and_size(self, field):
        """
        Get the offset and size of the given field from the beginning of the vector.

        Parameters
        ----------
        field : str

        Returns
        -------
        None|int
        """

        if field not in self._fields:
            return None

        out = 0
        for fld in self._fields:
            val = getattr(self, fld)
            if fld == field:
                return out, val
            else:
                out += val
        return None

    def get_dtype_components(self):
        """
        Gets the dtype components.

        Returns
        -------
        List[Tuple]
        """

        return [(entry, '>f8') for entry in self._fields]


class TOAParametersType(Serializable):
    """
    The TOA vector parameters.
    """

    _fields = ('DeltaTOA0', 'TOA_SS')
    _required = _fields
    # descriptors
    DeltaTOA0 = IntegerEnumDescriptor(
        'DeltaTOA0', (8, ), _required, strict=DEFAULT_STRICT, default_value=8,
        docstring='The size of the DeltaTOA0 field')  # type: int
    TOA_SS = IntegerEnumDescriptor(
        'TOA_SS', (8, ), _required, strict=DEFAULT_STRICT, default_value=8,
        docstring='The size of the TOA_SS field')  # type: int

    def __init__(self, DeltaTOA0=8, TOA_SS=8, **kwargs):
        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.DeltaTOA0 = DeltaTOA0
        self.TOA_SS = TOA_SS
        super(TOAParametersType, self).__init__(**kwargs)

    @staticmethod
    def get_size():
        """
        The size in bytes of this component of the vector.

        Returns
        -------
        int
        """

        return 16

    def get_position_offset_and_size(self, field):
        """
        Get the offset and size of the given field from the beginning of the vector.

        Parameters
        ----------
        field : str

        Returns
        -------
        None|(int, int)
        """

        if field not in self._fields:
            return None

        out = 0
        for fld in self._fields:
            val = getattr(self, fld)
            if fld == field:
                return out, val
            else:
                out += val
        return None

    def get_dtype_components(self):
        """
        Gets the dtype components.

        Returns
        -------
        List[Tuple]
        """

        return [(entry, '>f8') for entry in self._fields]


class VectorParametersType(Serializable):
    """
    The vector parameters sizes object.
    """

    _fields = (
        'TxTime', 'TxPos', 'RcvTime', 'RcvPos', 'SRPTime', 'SRPPos', 'AmpSF', 'TropoSRP',
        'FxParameters', 'TOAParameters')
    _required = (
        'TxTime', 'TxPos', 'RcvTime', 'RcvPos', 'SRPPos')
    _choice = ({'required': False, 'collection': ('FxParameters', 'TOAParameters')}, )
    # descriptors
    TxTime = IntegerEnumDescriptor(
        'TxTime', (8, ), _required, strict=DEFAULT_STRICT, default_value=8,
        docstring='The size of the TxTime field')  # type: int
    TxPos = IntegerEnumDescriptor(
        'TxPos', (24, ), _required, strict=DEFAULT_STRICT, default_value=8,
        docstring='The size of the TxPos field')  # type: int
    RcvTime = IntegerEnumDescriptor(
        'RcvTime', (8, ), _required, strict=DEFAULT_STRICT, default_value=8,
        docstring='The size of the RcvTime field')  # type: int
    RcvPos = IntegerEnumDescriptor(
        'RcvPos', (24, ), _required, strict=DEFAULT_STRICT, default_value=8,
        docstring='The size of the RcvPos field')  # type: int
    SRPTime = IntegerEnumDescriptor(
        'SRPTime', (8, ), _required, strict=DEFAULT_STRICT, default_value=None,
        docstring='The size of the SRPTime field')  # type: int
    SRPPos = IntegerEnumDescriptor(
        'SRPPos', (24, ), _required, strict=DEFAULT_STRICT, default_value=8,
        docstring='The size of the SRPPos field')  # type: int
    AmpSF = IntegerEnumDescriptor(
        'AmpSF', (8, ), _required, strict=DEFAULT_STRICT, default_value=None,
        docstring='The size of the AmpSF field')  # type: int
    TropoSRP = IntegerEnumDescriptor(
        'TropoSRP', (8, ), _required, strict=DEFAULT_STRICT, default_value=None,
        docstring='The size of the TropoSRP field')  # type: int
    FxParameters = SerializableDescriptor(
        'FxParameters', FxParametersType, _required, strict=DEFAULT_STRICT,
        docstring='The frequency parameters, only present when DomainType is '
                  'FX.')  # type: Union[None, FxParametersType]
    TOAParameters = SerializableDescriptor(
        'TOAParameters', FxParametersType, _required, strict=DEFAULT_STRICT,
        docstring='The TOA parameters, only present when DomainType is '
                  'TOA.')  # type: Union[None, TOAParametersType]

    def __init__(self, TxTime=8, TxPos=24, RcvTime=8, RcvPos=24, SRPTime=None, SRPPos=24,
                 AmpSF=None, TropoSRP=None, FxParameters=None, TOAParameters=None, **kwargs):
        """

        Parameters
        ----------
        TxTime : int
        TxPos : int
        RcvTime : int
        RcvPos : int
        SRPTime : None|int
        SRPPos : int
        AmpSF : None|int
        TropoSRP : None|int
        FxParameters : None|FxParametersType
        TOAParameters : None|TOAParametersType
        kwargs
        """

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.TxTime = TxTime
        self.TxPos = TxPos
        self.RcvTime = RcvTime
        self.RcvPos = RcvPos
        self.SRPTime = SRPTime
        self.SRPPos = SRPPos
        self.AmpSF = AmpSF
        self.TropoSRP = TropoSRP
        self.FxParameters = FxParameters
        self.TOAParameters = TOAParameters
        super(VectorParametersType, self).__init__(**kwargs)

    def get_size(self):
        """
        The size in bytes of this component of the vector.

        Returns
        -------
        int
        """

        out = 0
        for fld in self._fields:
            val = getattr(self, fld)
            if val is None:
                pass
            elif isinstance(val, int):
                out += val
            elif isinstance(val, (FxParametersType, TOAParametersType)):
                out += val.get_size()
            else:
                raise TypeError('Got unhandled type {}'.format(type(val)))
        return out

    def get_position_offset_and_size(self, field):
        """
        Get the offset and size of the given field from the beginning of the vector.

        Parameters
        ----------
        field : str

        Returns
        -------
        None|(int, int)
        """

        out = 0
        for fld in self._fields:
            val = getattr(self, fld)
            if fld == field:
                if val is not None:
                    return out, val
                else:
                    return None

            if val is None:
                pass
            elif isinstance(val, int):
                out += val
            elif isinstance(val, (FxParametersType, TOAParametersType)):
                res = val.get_position_offset_and_size(field)
                if res is not None:
                    return out+res[0], res[1]
                else:
                    out += val.get_size()
            else:
                raise TypeError('Got unhandled type {}'.format(type(val)))
        return None

    def get_vector_dtype(self):
        """
        Gets the dtype for the corresponding structured array for the full PVP array.

        Returns
        -------
        numpy.dtype
            This will be a compound dtype for a structured array.
        """

        the_type_info = []
        for fld in self._fields:
            val = getattr(self, fld)
            if val is None:
                continue
            if fld in ['FxParameters', 'TOAParameters']:
                the_type_info.extend(val.get_dtype_components())
            else:
                assert isinstance(val, int), 'CPHD 0.3 PVP field {} ' \
                                             'should be an integer, got {}'.format(fld, val)
                if val == 8:
                    the_type_info.append((fld, '>f8'))
                elif val == 24:
                    the_type_info.append((fld, '>f8', (3, )))
                else:
                    raise ValueError('Got unhandled value {} for CPHD 0.3 PVP field {}'.format(val, fld))
        return numpy.dtype(the_type_info)




from ..tre_elements import TREExtension, TREElement

__classification__ = "UNCLASSIFIED"
__author__ = "Thomas McCullough"


class MENSRA_155Type(TREElement):
    def __init__(self, value):
        super(MENSRA_155Type, self).__init__()
        self.add_field('CCRP_LOC', 's', 21, value)
        self.add_field('CCRP_ALT', 's', 6, value)
        self.add_field('OF_PC_R', 's', 7, value)
        self.add_field('OF_PC_A', 's', 7, value)
        self.add_field('COSGRZ', 's', 7, value)
        self.add_field('RGCCRP', 's', 7, value)
        self.add_field('RLMAP', 's', 1, value)
        self.add_field('CCRP_ROW', 's', 5, value)
        self.add_field('CCRP_COL', 's', 5, value)
        self.add_field('ACFT_LOC', 's', 21, value)
        self.add_field('ACFT_ALT', 's', 5, value)
        self.add_field('C_R_NC', 's', 7, value)
        self.add_field('C_R_EC', 's', 7, value)
        self.add_field('C_R_DC', 's', 7, value)
        self.add_field('C_AZ_NC', 's', 7, value)
        self.add_field('C_AZ_EC', 's', 7, value)
        self.add_field('C_AZ_DC', 's', 7, value)
        self.add_field('C_AL_NC', 's', 7, value)
        self.add_field('C_AL_EC', 's', 7, value)
        self.add_field('C_AL_DC', 's', 7, value)


class MENSRA_155(TREExtension):
    _tag_value = 'MENSRA'
    _data_type = MENSRA_155Type


class MENSRA_174Type(TREElement):
    def __init__(self, value):
        super(MENSRA_174Type, self).__init__()
        self.add_field('ACFT_LOC', 's', 21, value)
        self.add_field('ACFT_ALT', 's', 6, value)
        self.add_field('CCRP_LOC', 's', 21, value)
        self.add_field('CCRP_ALT', 's', 6, value)
        self.add_field('OF_PC_R', 's', 7, value)
        self.add_field('OF_PC_A', 's', 7, value)
        self.add_field('COSGRZ', 's', 7, value)
        self.add_field('RGCCRP', 's', 7, value)
        self.add_field('RLMAP', 's', 1, value)
        self.add_field('CCRP_ROW', 's', 5, value)
        self.add_field('CCRP_COL', 's', 5, value)
        self.add_field('C_R_NC', 's', 9, value)
        self.add_field('C_R_EC', 's', 9, value)
        self.add_field('C_R_DC', 's', 9, value)
        self.add_field('C_AZ_NC', 's', 9, value)
        self.add_field('C_AZ_EC', 's', 9, value)
        self.add_field('C_AZ_DC', 's', 9, value)
        self.add_field('C_AL_NC', 's', 9, value)
        self.add_field('C_AL_EC', 's', 9, value)
        self.add_field('C_AL_DC', 's', 9, value)


class MENSRA_174(TREExtension):
    _tag_value = 'MENSRA'
    _data_type = MENSRA_174Type


class MENSRA_185Type(TREElement):
    def __init__(self, value):
        super(MENSRA_185Type, self).__init__()
        self.add_field('ACFT_LOC', 's', 25, value)
        self.add_field('ACFT_ALT', 's', 6, value)
        self.add_field('CCRP_LOC', 's', 25, value)
        self.add_field('CCRP_ALT', 's', 6, value)
        self.add_field('OF_PC_R', 's', 7, value)
        self.add_field('OF_PC_A', 's', 7, value)
        self.add_field('COSGZ', 's', 7, value)
        self.add_field('RGCCRP', 's', 7, value)
        self.add_field('RLMAP', 's', 1, value)
        self.add_field('CCRP_ROW', 's', 5, value)
        self.add_field('CCRP_COL', 's', 5, value)
        self.add_field('C_R_NC', 's', 10, value)
        self.add_field('C_R_EC', 's', 10, value)
        self.add_field('C_R_DC', 's', 10, value)
        self.add_field('C_AZ_NC', 's', 9, value)
        self.add_field('C_AZ_EC', 's', 9, value)
        self.add_field('C_AZ_DC', 's', 9, value)
        self.add_field('C_AL_NC', 's', 9, value)
        self.add_field('C_AL_EC', 's', 9, value)
        self.add_field('C_AL_DC', 's', 9, value)
        self.add_field('TOTAL_TILES_COLS', 's', 3, value)
        self.add_field('TOTAL_TILES_ROWS', 's', 5, value)


class MENSRA_185(TREExtension):
    _tag_value = 'MENSRA'
    _data_type = MENSRA_185Type


class MENSRA(TREExtension):
    _tag_value = 'MENSRA'

    def __init__(self):
        raise ValueError(
            'Not to be implemented directly. '
            'Use of one MENSRA_155, MENSRA_174, or MENSRA_185')

    @classmethod
    def from_bytes(cls, value, start):
        """

        Parameters
        ----------
        value : bytes
        start : int

        Returns
        -------
        MENSRA_155|MENSRA_174|MENSRA_185
        """

        tag_value = value[start:start+6].decode('utf-8').strip()
        if tag_value != cls._tag_value:
            raise ValueError('tag value must be {}. Got {}'.format(cls._tag_value, tag_value))

        lng = int(value[start+6:start+11])
        if lng == 155:
            return MENSRA_155.from_bytes(value, start)
        elif lng == 174:
            return MENSRA_174.from_bytes(value, start)
        elif lng == 185:
            return MENSRA_185.from_bytes(value, start)
        else:
            raise ValueError('the data must be length 155, 174, or 185. Got {}'.format(lng))




from ..tre_elements import TREExtension, TREElement

__classification__ = "UNCLASSIFIED"
__author__ = "Thomas McCullough"


class SECTGAType(TREElement):
    def __init__(self, value):
        super(SECTGAType, self).__init__()
        self.add_field('SEC_ID', 's', 12, value)
        self.add_field('SEC_BE', 's', 15, value)
        self.add_field('RESVD001', 's', 1, value)


class SECTGA(TREExtension):
    _tag_value = 'SECTGA'
    _data_type = SECTGAType




from ..tre_elements import TREExtension, TREElement

__classification__ = "UNCLASSIFIED"
__author__ = "Thomas McCullough"


class MAP(TREElement):
    def __init__(self, value):
        super(MAP, self).__init__()
        self.add_field('MAP', 's', 21, value)


class CORSEG(TREElement):
    def __init__(self, value):
        super(CORSEG, self).__init__()
        self.add_field('CORSEG', 's', 21, value)
        self.add_field('TAUSEG', 's', 21, value)


class ERRCVG(TREElement):
    def __init__(self, value):
        super(ERRCVG, self).__init__()
        self.add_field('ERRCVG', 's', 21, value)


class ERRCOVDATA(TREElement):
    def __init__(self, value):
        super(ERRCOVDATA, self).__init__()
        self.add_field('NUMOPG', 'd', 2, value)
        self.add_loop(
            'ERRCVGs', int((self.NUMOPG+1)*self.NUMOPG/2), ERRCVG, value)
        self.add_field('TCDF', 'd', 1, value)
        self.add_field('NCSEG', 'd', 1, value)
        self.add_loop('CORSEGs', self.NCSEG, CORSEG, value)


class UCORSEGR(TREElement):
    def __init__(self, value):
        super(UCORSEGR, self).__init__()
        self.add_field('UCORSR', 's', 21, value)
        self.add_field('UTAUSR', 's', 21, value)


class UCORSEGC(TREElement):
    def __init__(self, value):
        super(UCORSEGC, self).__init__()
        self.add_field('UCORSC', 's', 21, value)
        self.add_field('UTAUSC', 's', 21, value)


class RSMECAType(TREElement):
    def __init__(self, value):
        super(RSMECAType, self).__init__()
        self.add_field('IID', 's', 80, value)
        self.add_field('EDITION', 's', 40, value)
        self.add_field('TID', 's', 40, value)
        self.add_field('INCLIC', 's', 1, value)
        self.add_field('INCLUC', 's', 1, value)

        if self.INCLIC == 'Y':
            self.add_field('NPAR', 'd', 2, value)
            self.add_field('NPARO', 'd', 2, value)
            self.add_field('IGN', 'd', 2, value)
            self.add_field('CVDATE', 's', 8, value)
            self.add_field('XUOL', 's', 21, value)
            self.add_field('YUOL', 's', 21, value)
            self.add_field('ZUOL', 's', 21, value)
            self.add_field('XUXL', 's', 21, value)
            self.add_field('XUYL', 's', 21, value)
            self.add_field('XUZL', 's', 21, value)
            self.add_field('YUXL', 's', 21, value)
            self.add_field('YUYL', 's', 21, value)
            self.add_field('YUZL', 's', 21, value)
            self.add_field('ZUXL', 's', 21, value)
            self.add_field('ZUYL', 's', 21, value)
            self.add_field('ZUZL', 's', 21, value)
            self.add_field('IR0', 's', 2, value)
            self.add_field('IRX', 's', 2, value)
            self.add_field('IRY', 's', 2, value)
            self.add_field('IRZ', 's', 2, value)
            self.add_field('IRXX', 's', 2, value)
            self.add_field('IRXY', 's', 2, value)
            self.add_field('IRXZ', 's', 2, value)
            self.add_field('IRYY', 's', 2, value)
            self.add_field('IRYZ', 's', 2, value)
            self.add_field('IRZZ', 's', 2, value)
            self.add_field('IC0', 's', 2, value)
            self.add_field('ICX', 's', 2, value)
            self.add_field('ICY', 's', 2, value)
            self.add_field('ICZ', 's', 2, value)
            self.add_field('ICXX', 's', 2, value)
            self.add_field('ICXY', 's', 2, value)
            self.add_field('ICXZ', 's', 2, value)
            self.add_field('ICYY', 's', 2, value)
            self.add_field('ICYZ', 's', 2, value)
            self.add_field('ICZZ', 's', 2, value)
            self.add_field('GX0', 's', 2, value)
            self.add_field('GY0', 's', 2, value)
            self.add_field('GZ0', 's', 2, value)
            self.add_field('GXR', 's', 2, value)
            self.add_field('GYR', 's', 2, value)
            self.add_field('GZR', 's', 2, value)
            self.add_field('GS', 's', 2, value)
            self.add_field('GXX', 's', 2, value)
            self.add_field('GXY', 's', 2, value)
            self.add_field('GXZ', 's', 2, value)
            self.add_field('GYX', 's', 2, value)
            self.add_field('GYY', 's', 2, value)
            self.add_field('GYZ', 's', 2, value)
            self.add_field('GZX', 's', 2, value)
            self.add_field('GZY', 's', 2, value)
            self.add_field('GZZ', 's', 2, value)
            self.add_loop('ERRCOVDATAs', self.IGN, ERRCOVDATA, value)
            self.add_loop('MAPMAT', int(self.NPAR * self.NPARO), MAP, value)

        if self.INCLUC == 'Y':
            self.add_field('URR', 's', 21, value)
            self.add_field('URC', 's', 21, value)
            self.add_field('UCC', 's', 21, value)
            self.add_field('UNCSR', 'd', 1, value)
            self.add_loop('UCORSEGRs', self.UNCSR, UCORSEGR, value)
            self.add_field('UNCSC', 'd', 1, value)
            self.add_loop('UCORSEGCs', self.UNCSC, UCORSEGC, value)


class RSMECA(TREExtension):
    _tag_value = 'RSMECA'
    _data_type = RSMECAType




from ..tre_elements import TREExtension, TREElement

__classification__ = "UNCLASSIFIED"
__author__ = "Thomas McCullough"


class VTGT(TREElement):
    def __init__(self, value):
        super(VTGT, self).__init__()
        self.add_field('TGLOC', 's', 21, value)
        self.add_field('TGRDV', 's', 4, value)
        self.add_field('TGGSP', 's', 3, value)
        self.add_field('TGHEA', 's', 3, value)
        self.add_field('TGSIG', 's', 2, value)
        self.add_field('TGCAT', 's', 1, value)


class MTIRPAType(TREElement):
    def __init__(self, value):
        super(MTIRPAType, self).__init__()
        self.add_field('DESTP', 's', 2, value)
        self.add_field('MTPID', 's', 3, value)
        self.add_field('PCHNO', 's', 4, value)
        self.add_field('WAMFN', 's', 5, value)
        self.add_field('WAMBN', 's', 1, value)
        self.add_field('UTC', 's', 8, value)
        self.add_field('SQNTA', 's', 5, value)
        self.add_field('COSGZ', 's', 7, value)
        self.add_field('NVTGT', 'd', 3, value)
        self.add_loop('VTGTs', self.NVTGT, VTGT, value)


class MTIRPA(TREExtension):
    _tag_value = 'MTIRPA'
    _data_type = MTIRPAType




from ..tre_elements import TREExtension, TREElement

__classification__ = "UNCLASSIFIED"
__author__ = "Thomas McCullough"


class PIAPEAType(TREElement):
    def __init__(self, value):
        super(PIAPEAType, self).__init__()
        self.add_field('LASTNME', 's', 28, value)
        self.add_field('FIRSTNME', 's', 28, value)
        self.add_field('MIDNME', 's', 28, value)
        self.add_field('DOB', 's', 6, value)
        self.add_field('ASSOCTRY', 's', 2, value)


class PIAPEA(TREExtension):
    _tag_value = 'PIAPEA'
    _data_type = PIAPEAType



"""
Sub-aperture processing methods.
"""

__author__ = 'Thomas McCullough'
__classification__ = "UNCLASSIFIED"


import logging
from typing import Union, Generator, Tuple, List, Optional, Sequence

import numpy
from scipy.constants import speed_of_light

from sarpy.processing.sicd.fft_base import FFTCalculator, fft, ifft, fftshift, fft2_sicd, ifft2_sicd
from sarpy.io.general.slice_parsing import validate_slice_int, verify_slice
from sarpy.processing.sicd.normalize_sicd import DeskewCalculator
from sarpy.processing.ortho_rectify import OrthorectificationHelper, OrthorectificationIterator
from sarpy.visualization.remap import RemapFunction
from sarpy.io.complex.base import SICDTypeReader
from sarpy.io.complex.sicd_elements.SICD import SICDType

logger = logging.getLogger(__name__)

####################
# Module variables providing default values
_METHOD_VALUES = ('NORMAL', 'FULL', 'MINIMAL')


def frame_definition(
        array_size: int,
        frame_count: int = 9,
        aperture_fraction: float = 0.2,
        fill: Union[int, float] = 1,
        method: str = 'FULL') -> Tuple[List[Tuple[int, int]], int]:
    """
    Get the frame definition along the desired axis for subaperture processing.

    Parameters
    ----------
    array_size : int
        The size of the given array.
    frame_count : int
        The number of frames to calculate.
    aperture_fraction : float
        The relative size of each aperture window.
    fill : float|int
        The fft fill value.
    method : str
        The subaperture processing method, which must be one of
        `('NORMAL', 'FULL', 'MINIMAL')`.

    Returns
    -------
    frame_definition: List[Tuple[int, int]]
    output_resolution: int
    """

    method = method.upper()
    if method not in _METHOD_VALUES:
        raise ValueError('method must be one of {}, got {}'.format(_METHOD_VALUES, method))

    fill = float(fill)
    if fill < 0.9999999:
        raise ValueError('fill must be at least 1.0, got {}'.format(fill))

    # determine our functional array and processing sizes
    functional_array_size = array_size/fill
    left_edge = int(numpy.round(0.5*(array_size - functional_array_size)))
    processing_size = array_size - 2*left_edge
    # determine the (static) size of each sub-aperture
    subaperture_size = int(numpy.ceil(aperture_fraction*processing_size))
    # determine the step size
    step = 0 if frame_count == 1 else \
        int(numpy.floor((processing_size - subaperture_size)/float(frame_count-1)))

    if method == 'NORMAL':
        output_resolution = int(numpy.ceil(aperture_fraction*array_size))
    elif method == 'FULL':
        output_resolution = array_size
    elif method == 'MINIMAL':
        output_resolution = int(numpy.ceil(processing_size/float(frame_count)))
    else:
        raise ValueError('Got unhandled method {}'.format(method))

    frames = []
    start_offset = left_edge
    for i in range(frame_count):
        frames.append((start_offset, start_offset + subaperture_size))
        start_offset += step

    return frames, output_resolution


#####################################
# The sub-aperture processing methods

def _validate_input(array: numpy.ndarray) -> numpy.ndarray:
    if not isinstance(array, numpy.ndarray):
        raise TypeError('array must be a numpy array. Got type {}'.format(type(array)))
    if not numpy.iscomplexobj(array):
        raise ValueError('array must be a complex array, got dtype {}'.format(array.dtype))
    if array.ndim != 2:
        raise ValueError('array must be two dimensional. Got shape {}'.format(array.shape))
    return array


def _validate_dimension(dimension: int) -> int:
    dimension = int(dimension)
    if dimension not in (0, 1):
        raise ValueError('dimension must be 0 or 1, got {}'.format(dimension))
    return dimension


def subaperture_processing_array(
        array: numpy.ndarray,
        aperture_indices: Tuple[int, int],
        output_resolution: int,
        dimension: int = 0) -> numpy.ndarray:
    """
    Perform the sub-aperture processing on the given complex array data.

    Parameters
    ----------
    array : numpy.ndarray
        The complex array data. Dimension other than 2 is not supported.
    aperture_indices : Tuple[int, int]
        The start/stop indices for the subaperture processing.
    output_resolution : int
        The output resolution parameter.
    dimension : int
        The dimension along which to perform the sub-aperture processing. Must be
        one of 0 or 1.

    Returns
    -------
    numpy.ndarray
    """

    array = _validate_input(array)
    dimension = _validate_dimension(dimension)

    return subaperture_processing_phase_history(
        fftshift(fft(array, axis=dimension), axes=dimension),
        aperture_indices, output_resolution, dimension=dimension)


def subaperture_processing_phase_history(
        phase_array: numpy.ndarray,
        aperture_indices: Tuple[int, int],
        output_resolution: int,
        dimension: int = 0) -> numpy.ndarray:
    """
    Perform the sub-aperture processing on the given complex phase history data.

    Parameters
    ----------
    phase_array : numpy.ndarray
        The complex array data. Dimension other than 2 is not supported.
    aperture_indices : Tuple[int, int]
        The start/stop indices for the subaperture processing.
    output_resolution : int
        The output resolution parameter.
    dimension : int
        The dimension along which to perform the sub-aperture processing. Must be
        one of 0 or 1.

    Returns
    -------
    numpy.ndarray
    """

    phase_array = _validate_input(phase_array)
    dimension = _validate_dimension(dimension)

    if dimension == 0:
        return ifft(phase_array[aperture_indices[0]:aperture_indices[1], :], axis=0, n=output_resolution)
    else:
        return ifft(phase_array[:, aperture_indices[0]:aperture_indices[1]], axis=1, n=output_resolution)


class SubapertureCalculator(FFTCalculator):
    """
    Class for performing sub-aperture processing from a reader instance.

    It is important to note that full resolution is required for along the
    processing dimension, so sub-sampling along the processing dimension does
    not decrease the amount of data which must be fetched.
    """

    __slots__ = ('_frame_count', '_aperture_fraction', '_method', '_frame_definition')

    def __init__(
            self,
            reader: Union[str, SICDTypeReader],
            dimension: int = 0,
            index: int = 0,
            block_size: int = 10,
            frame_count: int = 9,
            aperture_fraction: float = 0.2,
            method: str = 'FULL'):
        """

        Parameters
        ----------
        reader : str|SICDTypeReader
            Input file path or reader object, which must be of sicd type.
        dimension : int
            The dimension over which to split the sub-aperture.
        index : int
            The sicd index to use.
        block_size : int
            The approximate processing block size to fetch, given in MB. The
            minimum value for use here will be 1.
        frame_count : int
            The number of frames to calculate.
        aperture_fraction : float
            The relative size of each aperture window.
        method : str
            The subaperture processing method, which must be one of
            `('NORMAL', 'FULL', 'MINIMAL')`.
        """

        self._frame_count = 9
        self._aperture_fraction = 0.2
        self._method = 'FULL'
        self._frame_definition = None
        super(SubapertureCalculator, self).__init__(
            reader, dimension=dimension, index=index, block_size=block_size)

        self.frame_count = frame_count
        self.aperture_fraction = aperture_fraction
        self.method = method

    @property
    def frame_count(self) -> int:
        """
        int: The frame count.
        """

        return self._frame_count

    @frame_count.setter
    def frame_count(self, value):
        value = int(value)
        if value < 1:
            raise ValueError('frame_count must be a positive integer.')
        self._frame_count = value

    @property
    def aperture_fraction(self) -> float:
        """
        float: The relative aperture fraction size.
        """
        return self._aperture_fraction

    @aperture_fraction.setter
    def aperture_fraction(self, value):
        value = float(value)
        if not (0 < value < 1):
            raise ValueError(
                'aperture_fraction must be in the range (0, 1), got {}'.format(value))
        self._aperture_fraction = value

    @property
    def method(self) -> str:
        """
        str: The subaperture method.
        """

        return self._method

    @method.setter
    def method(self, value):
        value = value.upper()
        if value not in _METHOD_VALUES:
            raise ValueError('method must be one of {}, got {}'.format(_METHOD_VALUES, value))
        self._method = value

    def _parse_frame_argument(self, the_frame):
        if the_frame is None:
            return numpy.arange(self.frame_count)
        elif isinstance(the_frame, slice):
            the_frame = verify_slice(the_frame, self.frame_count)
            return numpy.arange(the_frame.start, the_frame.stop, the_frame.step)
        elif isinstance(the_frame, int):
            return validate_slice_int(the_frame, self.frame_count)
        elif isinstance(the_frame, (list, tuple)):
            return self._parse_frame_argument(numpy.array(the_frame))
        elif isinstance(the_frame, numpy.ndarray):
            if not issubclass(the_frame.dtype.type, numpy.integer):
                raise ValueError(
                    'The last slice dimension was a numpy array of unsupported non-integer '
                    'dtype {}'.format(the_frame.dtype))
            if the_frame.ndim != 1:
                raise ValueError(
                    'The last slice dimension was a numpy array which was not one-dimensional, '
                    'which is of unsupported.')
            out = the_frame.copy()
            for i, entry in enumerate(out):
                if (entry <= -self.frame_count) or (entry >= self.frame_count):
                    raise ValueError(
                        'The last slice dimension was a numpy array, and entry {} has '
                        'value {}, which is not sensible for the '
                        'bound {}'.format(i, entry, self.frame_count))
                if entry < 0:
                    out[i] += self.frame_count
            return out
        else:
            raise TypeError(
                'The final slice dimension is of unsupported type {}'.format(type(the_frame)))

    def _parse_slicing(self, item) -> Tuple[slice, slice, Optional[int]]:
        row_range, col_range, the_frame = super(SubapertureCalculator, self)._parse_slicing(item)
        return row_range, col_range, self._parse_frame_argument(the_frame)

    def subaperture_generator(
            self,
            row_range: Union[slice, Tuple[int, int, int]],
            col_range: Union[slice, Tuple[int, int, int]],
            frames: Union[None, int, list, tuple, numpy.ndarray] = None) -> Generator[numpy.ndarray, None, None]:
        """
        Supplies a generator for the given row and column ranges and frames collection.
        **Note that this IGNORES the block_size parameter in fetching, and fetches the
        entire required block.**

        The full resolution data in the processing dimension is required, even if
        down-sampled by the row_range or col_range parameter.

        Parameters
        ----------
        row_range : slice|Tuple[int, int, int]
            The row range.
        col_range : slice|Tuple[int, int, int]
            The column range.
        frames : None|int|list|tuple|numpy.ndarray
            The frame or frame collection.

        Returns
        -------
        Generator[numpy.ndarray]
        """

        def get_dimension_details(
                the_range: Union[slice, Tuple[int, int, int]]) -> Tuple[Tuple[int, int, int], int, int]:
            if isinstance(the_range, Sequence):
                start, stop, step = the_range
            elif isinstance(the_range, slice):
                start = the_range.start
                stop = the_range.stop
                step = the_range.step
            else:
                raise TypeError('Got unexpected range input {}'.format(the_range))

            the_snip = -1 if step < 0 else 1
            t_full_range = (start, stop, the_snip)
            t_full_size = stop - start
            t_step = abs(step)
            return t_full_range, t_full_size, t_step

        if self._fill is None:
            raise ValueError('Unable to proceed unless the index and dimension are set.')

        frames = self._parse_frame_argument(frames)
        if isinstance(frames, int):
            frames = [frames, ]

        if isinstance(row_range, tuple):
            row_slice = slice(*row_range)
        else:
            row_slice = row_range
        if isinstance(col_range, tuple):
            col_slice = slice(*col_range)
        else:
            col_slice = row_range

        if self.dimension == 0:
            # determine the full resolution block of data to fetch
            this_row_range, full_size, step = get_dimension_details(row_range)
            # fetch the necessary data block
            data = self.reader[(slice(*this_row_range), col_slice, self.index)]
        else:
            # determine the full resolution block of data to fetch
            this_col_range, full_size, step = get_dimension_details(col_range)
            # fetch the necessary data block, and transform to phase space
            data = self.reader[(row_slice, slice(*this_col_range), self.index)]
        # handle any nonsense data as 0
        data[~numpy.isfinite(data)] = 0
        # transform the data to phase space
        data = fftshift(fft(data, axis=self.dimension), axes=self.dimension)
        # define our frame collection
        frame_collection, output_resolution = frame_definition(
            full_size, frame_count=self.frame_count, aperture_fraction=self.aperture_fraction,
            fill=self.fill, method=self.method)
        # iterate over frames and generate the results
        for frame_index in frames:
            frame_def = frame_collection[int(frame_index)]
            this_subap_data = subaperture_processing_phase_history(
                data, frame_def, output_resolution=output_resolution, dimension=self.dimension)
            if step == 1:
                yield this_subap_data
            elif self.dimension == 0:
                yield this_subap_data[::step, :]
            else:
                yield this_subap_data[:, ::step]

    def _prepare_output(
            self,
            row_range: Union[slice, Tuple[int, int, int]],
            col_range: Union[slice, Tuple[int, int, int]],
            frames: Union[None, int, list, tuple, numpy.ndarray] = None) -> numpy.ndarray:
        row_start, row_stop, row_step = row_range if isinstance(row_range, tuple) else \
            (row_range.start, row_range.stop, row_range.step)
        if row_stop is None:
            if row_step > 0:
                raise ValueError('Got unexpected row_range {}'.format(row_range))
            row_stop = -1

        col_start, col_stop, col_step = col_range if isinstance(col_range, tuple) else \
            (col_range.start, col_range.stop, col_range.step)
        if col_stop is None:
            if col_step > 0:
                raise ValueError('Got unexpected col_range {}'.format(col_range))
            col_stop = -1

        row_count = int((row_stop - row_start)/float(row_step))
        col_count = int((col_stop - col_start)/float(col_step))
        if frames is None or len(frames) == 1:
            out_size = (row_count, col_count)
        else:
            out_size = (row_count, col_count, len(frames))
        return numpy.zeros(out_size, dtype=numpy.complex64)

    def __getitem__(self, item) -> numpy.ndarray:
        """
        Fetches the csi data based on the input slice. Slicing in the final
        dimension using an integer, slice, or integer array is supported. Note
        that this could easily be memory intensive, and should be used with
        some care.

        Parameters
        ----------
        item

        Returns
        -------
        numpy.ndarray
        """

        if self._fill is None:
            raise ValueError('Unable to proceed unless the index and dimension are set.')

        # parse the slicing to ensure consistent structure
        row_range, col_range, frames = self._parse_slicing(item)
        if isinstance(frames, int):
            frames = [frames, ]

        if self.dimension == 0:
            column_block_size = self.get_fetch_block_size(row_range.start, row_range.stop)
            # get our block definitions
            column_blocks, result_blocks = self.extract_blocks(col_range, column_block_size)
            if column_blocks == 1 and len(frames) == 1:
                # no need to prepare output, which will take twice the memory, so just return
                out = self.subaperture_generator(row_range, col_range, frames).__next__()
            else:
                out = self._prepare_output(row_range, col_range, frames=frames)
                # noinspection PyTypeChecker
                for this_column_range, result_range in zip(column_blocks, result_blocks):
                    generator = self.subaperture_generator(row_range, this_column_range, frames)
                    if len(frames) == 1:
                        out[:, result_range[0]:result_range[1]] = generator.__next__()
                    else:
                        for i, data in enumerate(generator):
                            out[:, result_range[0]:result_range[1], i] = data
        else:
            row_block_size = self.get_fetch_block_size(col_range.start, col_range.stop)
            # get our block definitions
            row_blocks, result_blocks = self.extract_blocks(row_range, row_block_size)
            if row_blocks == 1 and len(frames) == 1:
                out = self.subaperture_generator(row_range, col_range, frames).__next__()
            else:
                out = self._prepare_output(row_range, col_range, frames=frames)
                # noinspection PyTypeChecker
                for this_row_range, result_range in zip(row_blocks, result_blocks):
                    generator = self.subaperture_generator(this_row_range, col_range, frames)
                    if len(frames) == 1:
                        out[result_range[0]:result_range[1], :] = generator.__next__()
                    else:
                        for i, data in enumerate(generator):
                            out[result_range[0]:result_range[1], :, i] = data
        return out


class SubapertureOrthoIterator(OrthorectificationIterator):
    """
    An iterator class for the ortho-rectified subaperture processing.

    Iterating depth first requires the least fetching from the reader once for
    all frames. Otherwise, iterating requires redundantly fetching data once
    for each frame.

    It should be noted that fetching data is not time intensive if working using
    a local file (i.e. on your computer), but it may be if working using some
    kind of network file system.
    """

    __slots__ = ('_depth_first', '_this_frame', '_generator')

    def __init__(
            self,
            ortho_helper: OrthorectificationHelper,
            calculator: SubapertureCalculator,
            bounds: Union[None, numpy.ndarray, tuple, list] = None,
            remap_function: Optional[RemapFunction] = None,
            recalc_remap_globals: bool = False,
            depth_first: bool = True):
        """

        Parameters
        ----------
        ortho_helper : OrthorectificationHelper
        calculator : SubapertureCalculator
        bounds : None|numpy.ndarray|list|tuple
            The pixel bounds of the form `(min row, max row, min col, max col)`.
            This will default to the full image.
        remap_function : None|RemapFunction
            The remap function to apply, if desired.
        recalc_remap_globals : bool
            Only applies if a remap function is provided, should we recalculate
            any required global parameters? This will automatically happen if
            they are not already set.
        depth_first : bool
            If `True`, by image segment part then frame - this requires the least
            fetching from the reader, once across all frames. Otherwise, iteration
            will proceed by frames and then image segment - this requires more
            fetching from the reader, once per frame.
        """

        self._generator = None
        self._this_frame = None
        self._depth_first = bool(depth_first)

        if not isinstance(calculator, SubapertureCalculator):
            raise TypeError(
                'calculator must be an instance of SubapertureCalculator. \n\t'
                'Got type {}'.format(type(calculator)))
        super(SubapertureOrthoIterator, self).__init__(
            ortho_helper, calculator=calculator, bounds=bounds,
            remap_function=remap_function, recalc_remap_globals=recalc_remap_globals)

    @property
    def calculator(self) -> SubapertureCalculator:
        # noinspection PyTypeChecker
        return self._calculator

    def _depth_first_iteration(self) -> Tuple[numpy.ndarray, Tuple[int, int], int]:
        if not self._depth_first:
            raise ValueError('Requires depth_first = True')

        # determine our current state
        if self._this_index is None or self._this_frame is None:
            self._this_index = 0
            self._this_frame = 0
        else:
            self._this_frame += 1
            if self._this_frame >= self.calculator.frame_count:
                self._this_index += 1
                self._this_frame = 0
        # at this point, _this_index & _this_frame indicates which entry to return
        if self._this_index >= len(self._iteration_blocks):
            self._this_index = None  # reset the iteration scheme
            self._this_frame = None
            raise StopIteration()

        this_ortho_bounds, this_pixel_bounds = self._get_state_parameters()
        # accommodate for real pixel limits
        this_pixel_bounds = self._ortho_helper.get_real_pixel_bounds(this_pixel_bounds)
        if self._this_frame == 0:
            # set up the iterator from the calculator
            self._generator = self.calculator.subaperture_generator(
                (this_pixel_bounds[0], this_pixel_bounds[1], 1),
                (this_pixel_bounds[2], this_pixel_bounds[3], 1))

        logger.info(
            'Fetching orthorectified coordinate block ({}:{}, {}:{}) of ({}:{}) for frame {}'.format(
                this_ortho_bounds[0] - self.ortho_bounds[0], this_ortho_bounds[1] - self.ortho_bounds[0],
                this_ortho_bounds[2] - self.ortho_bounds[2], this_ortho_bounds[3] - self.ortho_bounds[2],
                self.ortho_bounds[1] - self.ortho_bounds[0], self.ortho_bounds[3] - self.ortho_bounds[2],
                self._this_frame))
        data = self._generator.__next__()
        ortho_data = self._get_orthorectified_version(this_ortho_bounds, this_pixel_bounds, data)
        start_indices = (this_ortho_bounds[0] - self.ortho_bounds[0],
                         this_ortho_bounds[2] - self.ortho_bounds[2])
        return ortho_data, start_indices, self._this_frame

    def _frame_first_iteration(self) -> Tuple[numpy.ndarray, Tuple[int, int], int]:
        if self._depth_first:
            raise ValueError('Requires depth_first = False')

        # determine our current state
        if self._this_index is None or self._this_frame is None:
            self._this_index = 0
            self._this_frame = 0
        else:
            self._this_index += 1
            if self._this_index >= len(self._iteration_blocks):
                self._this_frame += 1
                self._this_index = 0

        # at this point, _this_index & _this_frame indicates which entry to return
        if self._this_frame >= self.calculator.frame_count:
            raise StopIteration()

        # calculate our result
        this_ortho_bounds, this_pixel_bounds = self._get_state_parameters()
        # accommodate for real pixel limits
        this_pixel_bounds = self._ortho_helper.get_real_pixel_bounds(this_pixel_bounds)
        logger.info(
            'Fetching orthorectified coordinate block ({}:{}, {}:{}) of ({}:{}) for frame {}'.format(
                this_ortho_bounds[0] - self.ortho_bounds[0], this_ortho_bounds[1] - self.ortho_bounds[0],
                this_ortho_bounds[2] - self.ortho_bounds[2], this_ortho_bounds[3] - self.ortho_bounds[2],
                self.ortho_bounds[1] - self.ortho_bounds[0], self.ortho_bounds[3] - self.ortho_bounds[2],
                self._this_frame))

        data = self.calculator[
               this_pixel_bounds[0]:this_pixel_bounds[1],
               this_pixel_bounds[2]:this_pixel_bounds[3],
               self._this_frame]
        ortho_data = self._get_orthorectified_version(this_ortho_bounds, this_pixel_bounds, data)
        start_indices = (this_ortho_bounds[0] - self.ortho_bounds[0],
                         this_ortho_bounds[2] - self.ortho_bounds[2])
        return ortho_data, start_indices, self._this_frame

    def __next__(self) -> Tuple[numpy.ndarray, Tuple[int, int], int]:
        """
        Get the next iteration of ortho-rectified data.

        Returns
        -------
        data: numpy.ndarray
        normalized_indices: Tuple[int, int]
            The (normalized) indices (start_row, start_col) for this section of
            data, relative to overall output shape
        frame: int
            The frame index.
        """

        # NB: this is the Python 3 pattern for iteration

        if self._depth_first:
            return self._depth_first_iteration()
        else:
            return self._frame_first_iteration()

    def next(self) -> Tuple[numpy.ndarray, Tuple[int, int], int]:
        """
        Get the next iteration of ortho-rectified data.

        Returns
        -------
        data: numpy.ndarray
        normalized_indices: Tuple[int, int]
            The (normalized) indices (start_row, start_col) for this section of
            data, relative to overall output shape
        frame: int
            The frame index.
        """

        # NB: this is the Python 2 pattern for iteration
        return self.__next__()


class ApertureFilter(object):
    """
    This is a calculator for filtering SAR imagery using a subregion of complex
    fft data over a full resolution subregion of the original SAR data.

    This is largely intended as a helper function for the aperture GUI tool,
    but is included here because it may have wider applicability.
    """

    __slots__ = (
        '_deskew_calculator', '_sub_image_bounds', '_normalized_phase_history')

    def __init__(
            self,
            reader: SICDTypeReader,
            dimension: int = 1,
            index: int = 0,
            apply_deskew: bool = True,
            apply_deweighting: bool = False):
        """

        Parameters
        ----------
        reader : SICDTypeReader
        dimension : int
        index : int
        apply_deskew : bool
        apply_deweighting : bool
        """

        self._normalized_phase_history = None
        self._deskew_calculator = DeskewCalculator(
            reader, dimension=dimension, index=index, apply_deskew=apply_deskew, apply_deweighting=apply_deweighting)
        self._sub_image_bounds = None

    @property
    def apply_deskew(self) -> bool:
        """
        bool: Apply deskew to calculated value.
        """

        return self._deskew_calculator.apply_deskew

    @apply_deskew.setter
    def apply_deskew(self, value):
        self._deskew_calculator.apply_deskew = value
        self._set_normalized_phase_history()

    @property
    def apply_deweighting(self) -> bool:
        """
        bool: Apply deweighting to calculated values.
        """

        return self._deskew_calculator.apply_deweighting

    @apply_deweighting.setter
    def apply_deweighting(self, val):
        self._deskew_calculator.apply_deweighting = val
        self._set_normalized_phase_history()

    def _get_fft_complex_data(self, cdata: numpy.ndarray) -> numpy.ndarray:
        """
        Transform the complex image data to phase history data.

        Parameters
        ----------
        cdata : numpy.ndarray

        Returns
        -------
        numpy.ndarray
        """

        return fftshift(fft2_sicd(cdata, self.sicd))

    def _get_fft_phase_data(self, ph_data: numpy.ndarray) -> numpy.ndarray:
        """
        Transforms the phase history data to complex image data.

        Parameters
        ----------
        ph_data : numpy.ndarray

        Returns
        -------
        numpy.ndarray
        """

        return ifft2_sicd(ph_data, self.sicd)

    @property
    def sicd(self) -> SICDType:
        """
        SICDType: The associated SICD structure.
        """

        return self._deskew_calculator.sicd

    @property
    def dimension(self) -> int:
        """
        int: The processing dimension.
        """

        return self._deskew_calculator.dimension

    @property
    def data_size(self) -> Optional[Tuple[int, int]]:
        """
        None|(int, int): The feasible data size
        """

        if self._sub_image_bounds is None:
            return None
        row_bounds, col_bounds = self._sub_image_bounds
        return int(row_bounds[1] - row_bounds[0]), int(col_bounds[1] - col_bounds[0])

    @dimension.setter
    def dimension(self, val):
        """
        Parameters
        ----------
        val : int

        Returns
        -------
        None
        """

        self._deskew_calculator.dimension = val
        self._set_normalized_phase_history()

    @property
    def flip_x_axis(self) -> bool:
        try:
            return self.sicd.SCPCOA.SideOfTrack == "L"
        except AttributeError:
            return False

    @property
    def sub_image_bounds(self) -> Tuple[Tuple[int, int], ...]:
        """
        Tuple[Tuple[int, int], ...]: The sub-image bounds used for processing.
        """

        return self._sub_image_bounds

    def set_sub_image_bounds(
            self,
            row_bounds: Optional[Tuple[int, int]],
            col_bounds: Optional[Tuple[int, int]]):
        """
        Sets the full range bounds for the phase history calculation. This subsequently
        sets the `normalized_phase_history` value.

        Parameters
        ----------
        row_bounds : None|Tuple[int, int]
            Of the form `(start row, end row)`.
        col_bounds : None|Tuple[int, int]
            Of the form `(start column, end column)`.

        Returns
        -------
        None
        """

        def validate_entry(entry):
            if len(entry) != 2:
                raise ValueError('Bounds must have length 2. Got {}'.format(entry))
            out = (int(entry[0]), int(entry[1]))
            if out[0] >= out[1]:
                raise ValueError('Bounds must have int(bound[0]) < int(bound[1]). Got {}'.format(entry))
            return out

        if row_bounds is None or col_bounds is None:
            self._sub_image_bounds = None
        else:
            self._sub_image_bounds = (validate_entry(row_bounds), validate_entry(col_bounds))
        self._set_normalized_phase_history()

    @property
    def normalized_phase_history(self) -> Optional[numpy.ndarray]:
        """
        None|numpy.ndarray: The normalized phase history
        """

        return self._normalized_phase_history

    def _set_normalized_phase_history(self) -> None:
        """
        Sets the normalized phase history.

        Returns
        -------
        None
        """

        if self._sub_image_bounds is None:
            self._normalized_phase_history = None
            return

        row_bounds, col_bounds = self._sub_image_bounds
        underlying_size = self._deskew_calculator.data_size
        if row_bounds[0] < 0 or row_bounds[1] > underlying_size[0]:
            raise ValueError(
                'Desired row_bounds given as {}, and underlying data size is {}'.format(row_bounds, underlying_size))
        if col_bounds[0] < 0 or col_bounds[1] > underlying_size[1]:
            raise ValueError(
                'Desired col_bounds given as {}, and underlying data size is {}'.format(col_bounds, underlying_size))

        deskewed_data = self._deskew_calculator[row_bounds[0]:row_bounds[1], col_bounds[0]:col_bounds[1]]
        self._normalized_phase_history = self._get_fft_complex_data(deskewed_data)

    @property
    def polar_angles(self) -> numpy.ndarray:
        angle_width = (1 / self.sicd.Grid.Col.SS) / self.sicd.Grid.Row.KCtr
        if self.sicd.Grid.Col.KCtr:
            angle_ctr = self.sicd.Grid.Col.KCtr
        else:
            angle_ctr = 0
        angle_limits = angle_ctr + numpy.array([-1, 1]) * angle_width / 2
        if self.flip_x_axis:
            angle_limits = angle_limits[1], angle_limits[0]
        angles = numpy.linspace(angle_limits[0], angle_limits[1], self.normalized_phase_history.shape[1])
        return numpy.rad2deg(numpy.arctan(angles))

    @property
    def frequencies(self) -> numpy.ndarray:
        """
        This returns the subaperture frequencies in units of GHz

        Returns
        -------
        numpy.array
        """

        freq_width = (1 / self.sicd.Grid.Row.SS) * (speed_of_light / 2)
        freq_ctr = self.sicd.Grid.Row.KCtr * (speed_of_light / 2)
        freq_limits = freq_ctr + (numpy.array([-1, 1]) * freq_width / 2)
        if self.sicd.PFA:
            freq_limits = freq_limits / self.sicd.PFA.SpatialFreqSFPoly[0]
        freq_limits = freq_limits/1e9
        frequencies = numpy.linspace(freq_limits[1], freq_limits[0], self.normalized_phase_history.shape[0])
        return frequencies

    def __getitem__(self, item) -> Optional[numpy.ndarray]:
        if self.normalized_phase_history is None:
            return None
        filtered_cdata = numpy.zeros(self.normalized_phase_history.shape, dtype='complex64')
        filtered_cdata[item] = self.normalized_phase_history[item]
        # do the inverse transform of this sampled portion
        return self._get_fft_phase_data(filtered_cdata)




from ..tre_elements import TREExtension, TREElement

__classification__ = "UNCLASSIFIED"
__author__ = "Thomas McCullough"


class ASTORAType(TREElement):
    def __init__(self, value):
        super(ASTORAType, self).__init__()
        self.add_field('IMG_TOTAL_ROWS', 's', 6, value)
        self.add_field('IMG_TOTAL_COLS', 's', 6, value)
        self.add_field('IMG_INDEX_ROW', 's', 6, value)
        self.add_field('IMG_INDEX_COL', 's', 6, value)
        self.add_field('GEOID_OFFSET', 's', 7, value)
        self.add_field('ALPHA_0', 's', 16, value)
        self.add_field('K_L', 's', 2, value)
        self.add_field('C_M', 's', 15, value)
        self.add_field('AC_ROLL', 's', 16, value)
        self.add_field('AC_PITCH', 's', 16, value)
        self.add_field('AC_YAW', 's', 16, value)
        self.add_field('AC_TRACK_HEADING', 's', 16, value)
        self.add_field('AP_ORIGIN_X', 's', 13, value)
        self.add_field('AP_ORIGIN_Y', 's', 13, value)
        self.add_field('AP_ORIGIN_Z', 's', 13, value)
        self.add_field('AP_DIR_X', 's', 16, value)
        self.add_field('AP_DIR_Y', 's', 16, value)
        self.add_field('AP_DIR_Z', 's', 16, value)
        self.add_field('X_AP_START', 's', 12, value)
        self.add_field('X_AP_END', 's', 12, value)
        self.add_field('SS_ROW_SHIFT', 's', 4, value)
        self.add_field('SS_COL_SHIFT', 's', 4, value)
        self.add_field('U_HAT_X', 's', 16, value)
        self.add_field('U_HAT_Y', 's', 16, value)
        self.add_field('U_HAT_Z', 's', 16, value)
        self.add_field('V_HAT_X', 's', 16, value)
        self.add_field('V_HAT_Y', 's', 16, value)
        self.add_field('V_HAT_Z', 's', 16, value)
        self.add_field('N_HAT_X', 's', 16, value)
        self.add_field('N_HAT_Y', 's', 16, value)
        self.add_field('N_HAT_Z', 's', 16, value)
        self.add_field('ETA_0', 's', 16, value)
        self.add_field('SIGMA_SM', 's', 13, value)
        self.add_field('SIGMA_SN', 's', 13, value)
        self.add_field('S_OFF', 's', 10, value)
        self.add_field('RN_OFFSET', 's', 12, value)
        self.add_field('R_SCL', 's', 16, value)
        self.add_field('R_NAV', 's', 16, value)
        self.add_field('R_SC_EXACT', 's', 16, value)
        self.add_field('C_SC_X', 's', 16, value)
        self.add_field('C_SC_Y', 's', 16, value)
        self.add_field('C_SC_Z', 's', 16, value)
        self.add_field('K_HAT_X', 's', 16, value)
        self.add_field('K_HAT_Y', 's', 16, value)
        self.add_field('K_HAT_Z', 's', 16, value)
        self.add_field('L_HAT_X', 's', 16, value)
        self.add_field('L_HAT_Y', 's', 16, value)
        self.add_field('L_HAT_Z', 's', 16, value)
        self.add_field('P_Z', 's', 16, value)
        self.add_field('THETA_C', 's', 16, value)
        self.add_field('ALPHA_SL', 's', 16, value)
        self.add_field('SIGMA_TC', 's', 16, value)


class ASTORA(TREExtension):
    _tag_value = 'ASTORA'
    _data_type = ASTORAType




from ..tre_elements import TREExtension, TREElement

__classification__ = "UNCLASSIFIED"
__author__ = "Thomas McCullough"


class USE00AType(TREElement):
    def __init__(self, value):
        super(USE00AType, self).__init__()
        self.add_field('ANGLE_TO_NORTH', 's', 3, value)
        self.add_field('MEAN_GSD', 's', 5, value)
        self.add_field('RSRVD01', 's', 1, value)
        self.add_field('DYNAMIC_RANGE', 's', 5, value)
        self.add_field('RSRVD02', 's', 3, value)
        self.add_field('RSRVD03', 's', 1, value)
        self.add_field('RSRVD04', 's', 3, value)
        self.add_field('OBL_ANG', 's', 5, value)
        self.add_field('ROLL_ANG', 's', 6, value)
        self.add_field('RSRVD05', 's', 12, value)
        self.add_field('RSRVD06', 's', 15, value)
        self.add_field('RSRVD07', 's', 4, value)
        self.add_field('RSRVD08', 's', 1, value)
        self.add_field('RSRVD09', 's', 3, value)
        self.add_field('RSRVD10', 's', 1, value)
        self.add_field('RSRVD11', 's', 1, value)
        self.add_field('N_REF', 's', 2, value)
        self.add_field('REV_NUM', 's', 5, value)
        self.add_field('N_SEG', 's', 3, value)
        self.add_field('MAX_LP_SEG', 's', 6, value)
        self.add_field('RSRVD12', 's', 6, value)
        self.add_field('RSRVD13', 's', 6, value)
        self.add_field('SUN_EL', 's', 5, value)
        self.add_field('SUN_AZ', 's', 5, value)


class USE00A(TREExtension):
    _tag_value = 'USE00A'
    _data_type = USE00AType




from ..tre_elements import TREExtension, TREElement

__classification__ = "UNCLASSIFIED"
__author__ = "Thomas McCullough"


class CODE(TREElement):
    def __init__(self, value):
        super(CODE, self).__init__()
        self.add_field('CODE_LEN', 'd', 1, value)
        self.add_field('CODE', 's', self.CODE_LEN, value)
        self.add_field('EQTYPE', 's', 1, value)
        self.add_field('ESURN_LEN', 'd', 2, value)
        self.add_field('ESURN', 's', self.ESURN_LEN, value)
        self.add_field('DETAIL_LEN', 'd', 5, value)
        if self.DETAIL_LEN > 0:
            self.add_field('DETAIL_CMPR', 's', 1, value)
            self.add_field('DETAIL', 's', self.DETAIL_LEN, value)


class CCINFAType(TREElement):
    def __init__(self, value):
        super(CCINFAType, self).__init__()
        self.add_field('NUMCODE', 'd', 3, value)
        self.add_loop('CODEs', self.NUMCODE, CODE, value)


class CCINFA(TREExtension):
    _tag_value = 'CCINFA'
    _data_type = CCINFAType




from ..tre_elements import TREExtension, TREElement

__classification__ = "UNCLASSIFIED"
__author__ = "Thomas McCullough"


class IPCOM(TREElement):
    def __init__(self, value):
        super(IPCOM, self).__init__()
        self.add_field('IPCOM', 's', 80, value)


class EVENT(TREElement):
    def __init__(self, value):
        super(EVENT, self).__init__()
        self.add_field('PDATE', 's', 14, value)
        self.add_field('PSITE', 's', 10, value)
        self.add_field('PAS', 's', 10, value)
        self.add_field('NIPCOM', 'd', 1, value)
        self.add_loop('IPCOMs', self.NIPCOM, IPCOM, value)
        self.add_field('IBPP', 's', 2, value)
        self.add_field('IPVTYPE', 's', 3, value)
        self.add_field('INBWC', 's', 10, value)
        self.add_field('DISP_FLAG', 's', 1, value)
        self.add_field('ROT_FLAG', 's', 1, value)
        if self.ROT_FLAG == '1':
            self.add_field('ROT_ANGLE', 's', 8, value)
        self.add_field('ASYM_FLAG', 's', 1, value)
        if self.ASYM_FLAG == '1':
            self.add_field('ZOOMROW', 's', 7, value)
            self.add_field('ZOOMCOL', 's', 7, value)
        self.add_field('PROJ_FLAG', 's', 1, value)
        self.add_field('SHARP_FLAG', 's', 1, value)
        if self.SHARP_FLAG == '1':
            self.add_field('SHARPFAM', 's', 2, value)
            self.add_field('SHARPMEM', 's', 2, value)
        self.add_field('MAG_FLAG', 's', 1, value)
        if self.MAG_FLAG == '1':
            self.add_field('MAG_LEVEL', 's', 7, value)
        self.add_field('DRA_FLAG', 's', 1, value)
        if self.DRA_FLAG == '1':
            self.add_field('DRA_MULT', 's', 7, value)
            self.add_field('DRA_SUB', 's', 5, value)
        self.add_field('TTC_FLAG', 's', 1, value)
        if self.TTC_FLAG == '1':
            self.add_field('TTCFAM', 's', 2, value)
            self.add_field('TTCMEM', 's', 2, value)
        self.add_field('DEVLUT_FLAG', 's', 1, value)
        self.add_field('OBPP', 's', 2, value)
        self.add_field('OPVTYPE', 's', 3, value)
        self.add_field('OUTBWC', 's', 10, value)


class HISTOAType(TREElement):
    def __init__(self, value):
        super(HISTOAType, self).__init__()
        self.add_field('SYSTYPE', 's', 20, value)
        self.add_field('PC', 's', 12, value)
        self.add_field('PE', 's', 4, value)
        self.add_field('REMAP_FLAG', 's', 1, value)
        self.add_field('LUTID', 's', 2, value)
        self.add_field('NEVENTS', 'd', 2, value)
        self.add_loop('EVENTs', self.NEVENTS, EVENT, value)


class HISTOA(TREExtension):
    _tag_value = 'HISTOA'
    _data_type = HISTOAType



"""
Base structures for phase history readers and usage
"""

__classification__ = "UNCLASSIFIED"
__author__ = "Thomas McCullough"


from typing import Union, Tuple, Sequence, Dict, Optional

import numpy

from sarpy.io.general.base import BaseReader
from sarpy.io.general.data_segment import DataSegment
from sarpy.io.phase_history.cphd1_elements.CPHD import CPHDType as CPHDType1_0
from sarpy.io.phase_history.cphd0_3_elements.CPHD import CPHDType as CPHDType0_3


class CPHDTypeReader(BaseReader):
    """
    A class for common CPHD reading functionality.

    **Updated in version 1.3.0**
    """

    def __init__(
            self,
            data_segment: Union[None, DataSegment, Sequence[DataSegment]],
            cphd_meta: Union[None, CPHDType1_0, CPHDType0_3],
            close_segments: bool = True,
            delete_files: Union[None, str, Sequence[str]] = None):
        """

        Parameters
        ----------
        data_segment : None|DataSegment|Sequence[DataSegment]
        cphd_meta : None|CPHDType1_0|CPHDType0_3
            The CPHD metadata object
        close_segments : bool
            Call segment.close() for each data segment on reader.close()?
        delete_files : None|Sequence[str]
            Any temp files which should be cleaned up on reader.close()?
            This will occur after closing segments.
        """

        if cphd_meta is None:
            self._cphd_meta = None
        elif isinstance(cphd_meta, (CPHDType1_0, CPHDType0_3)):
            self._cphd_meta = cphd_meta
        else:
            raise TypeError(
                'The cphd_meta must be of type CPHDType, got `{}`'.format(type(cphd_meta)))

        BaseReader.__init__(
            self, data_segment, reader_type='CPHD', close_segments=close_segments, delete_files=delete_files)

    @property
    def cphd_meta(self) -> Union[None, CPHDType1_0, CPHDType0_3]:
        """
        None|CPHDType1_0|CPHDType0_3: the cphd meta_data.
        """

        return self._cphd_meta

    def read_support_array(
            self,
            index: Union[int, str],
            *ranges: Sequence[Union[None, int, Tuple[int, ...], slice]]) -> numpy.ndarray:
        """
        Read the support array.

        Parameters
        ----------
        index : int|str
            The support array integer index.
        ranges : Sequence[None|int|Tuple[int, ...]|slice]
            The slice definition appropriate for support array usage.

        Returns
        -------
        numpy.ndarray

        Raises
        ------
        TypeError
            If called on a reader which doesn't support this.
        """

        raise TypeError('Class {} does not provide support arrays'.format(type(self)))

    def read_support_block(self) -> Dict[str, numpy.ndarray]:
        """
        Reads the entirety of support block(s).

        Returns
        -------
        Dict[str, numpy.ndarray]
            Dictionary of `numpy.ndarray` containing the support arrays.
        """

        raise TypeError('Class {} does not provide support arrays'.format(type(self)))

    def read_pvp_variable(
            self,
            variable: str,
            index: Union[int, str],
            the_range: Union[None, int, Tuple[int, ...], slice] = None) -> Optional[numpy.ndarray]:
        """
        Read the vector parameter for the given `variable` and CPHD channel.

        Parameters
        ----------
        variable : str
        index : int|str
            The channel index or identifier.
        the_range : None|int|Tuple[int, ...]|slice
            The indices for the vector parameter. `None` returns all,
            a integer returns the single value at that location, otherwise
            the input determines a slice.

        Returns
        -------
        None|numpy.ndarray
            This will return None if there is no such variable, otherwise the data.
        """

        raise NotImplementedError

    def read_pvp_array(
            self,
            index: Union[int, str],
            the_range: Union[None, int, Tuple[int, ...], slice] = None) -> numpy.ndarray:
        """
        Read the PVP array from the requested channel.

        Parameters
        ----------
        index : int|str
            The support array integer index (of cphd.Data.Channels list) or identifier.
        the_range : None|int|Tuple[int, ...]|slice
            The indices for the vector parameter. `None` returns all,
            a integer returns the single value at that location, otherwise
            the input determines a slice.

        Returns
        -------
        pvp_array : numpy.ndarray
        """

        raise NotImplementedError

    def read_pvp_block(self) -> Dict[Union[int, str], numpy.ndarray]:
        """
        Reads the entirety of the PVP block(s).

        Returns
        -------
        Dict[Union[int, str], numpy.ndarray]
            Dictionary containing the PVP arrays.
        """

        raise NotImplementedError

    def read_signal_block(self) -> Dict[Union[int, str], numpy.ndarray]:
        """
        Reads the entirety of signal block(s), with data formatted as complex64
        (after accounting for AmpSF).

        Returns
        -------
        Dict[Union[int, str], numpy.ndarray]
            Dictionary of `numpy.ndarray` containing the signal arrays.
        """

        raise NotImplementedError

    def read_signal_block_raw(self) -> Dict[Union[int, str], numpy.ndarray]:
        """
        Reads the entirety of signal block(s), with data formatted in file
        storage format (no converting to complex, no consideration of AmpSF).

        Returns
        -------
        Dict[Union[int, str], numpy.ndarray]
            Dictionary of `numpy.ndarray` containing the signal arrays.
        """

        raise NotImplementedError



"""
The ProductDisplayType definition for SIDD 1.0.
"""

__classification__ = "UNCLASSIFIED"
__author__ = "Thomas McCullough"

import logging
from typing import Union
from collections import OrderedDict

import numpy

from sarpy.io.product.sidd2_elements.base import DEFAULT_STRICT, FLOAT_FORMAT
from sarpy.io.xml.base import Serializable, Arrayable, ParametersCollection, \
    create_text_node, create_new_node, get_node_value, find_first_child
from sarpy.io.xml.descriptors import SerializableDescriptor, IntegerDescriptor, \
    FloatDescriptor, StringDescriptor, StringEnumDescriptor, ParametersDescriptor


logger = logging.getLogger(__name__)


class ColorDisplayRemapType(Serializable, Arrayable):
    """
    LUT-base remap indicating that the color display is done through index-based color.
    """

    __slots__ = ('_remap_lut', )

    def __init__(self, RemapLUT=None, **kwargs):
        """

        Parameters
        ----------
        RemapLUT : numpy.ndarray|list|tuple
        kwargs
        """

        self._remap_lut = None
        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.RemapLUT = RemapLUT
        super(ColorDisplayRemapType, self).__init__(**kwargs)

    @property
    def RemapLUT(self):
        """
        numpy.ndarray: the two dimensional (:math:`N \times 3`) look-up table, where the dtype must be
        `uint8` or `uint16`. The first dimension should correspond to entries (i.e. size of the lookup table), and the
        second dimension must have size 3 and corresponds to `RGB` bands.
        """

        return self._remap_lut

    @RemapLUT.setter
    def RemapLUT(self, value):
        if value is None:
            self._remap_lut = None
            return
        if isinstance(value, (tuple, list)):
            value = numpy.array(value, dtype=numpy.uint8)
        if not isinstance(value, numpy.ndarray) or value.dtype.name not in ('uint8', 'uint16'):
            raise ValueError(
                'RemapLUT for class ColorDisplayRemapType must be a numpy.ndarray of dtype uint8 or uint16.')
        if value.ndim != 2 and value.shape[1] != 3:
            raise ValueError('RemapLUT for class ColorDisplayRemapType must be an N x 3 array.')
        self._remap_lut = value

    @property
    def size(self):
        """
        int: the size of the lookup table
        """
        if self._remap_lut is None:
            return 0
        else:
            return self._remap_lut.shape[0]

    def __len__(self):
        if self._remap_lut is None:
            return 0
        return self._remap_lut.shape[0]

    def __getitem__(self, item):
        return self._remap_lut[item]

    @classmethod
    def from_array(cls, array):
        """
        Create from the lookup table array.

        Parameters
        ----------
        array: numpy.ndarray|list|tuple
            Must be two-dimensional. If not a numpy.ndarray, this will be naively
            interpreted as `uint8`.

        Returns
        -------
        LUTInfoType
        """

        return cls(RemapLUT=array)

    def get_array(self, dtype=numpy.uint8):
        """
        Gets **a copy** of the coefficient array of specified data type.

        Parameters
        ----------
        dtype : str|numpy.dtype|numpy.number
            numpy data type of the return

        Returns
        -------
        numpy.ndarray
            the lookup table array
        """

        return numpy.array(self._remap_lut, dtype=dtype)

    @classmethod
    def from_node(cls, node, xml_ns, ns_key=None, kwargs=None):
        lut_key = cls._child_xml_ns_key.get('RemapLUT', ns_key)
        lut_node = find_first_child(node, 'RemapLUT', xml_ns, lut_key)
        if lut_node is not None:
            dim1 = int(lut_node.attrib['size'])
            dim2 = 3
            arr = numpy.zeros((dim1, dim2), dtype=numpy.uint16)
            entries = get_node_value(lut_node).split()
            i = 0
            for entry in entries:
                if len(entry) == 0:
                    continue
                sentry = entry.split(',')
                if len(sentry) != 3:
                    logger.error(
                        'Parsing RemapLUT is likely compromised.\n\t'
                        'Got entry {}, which we are skipping.'.format(entry))
                    continue
                arr[i, :] = [int(el) for el in entry]
                i += 1
            if numpy.max(arr) < 256:
                arr = numpy.asarray(arr, dtype=numpy.uint8)
            return cls(RemapLUT=arr)
        return cls()

    def to_node(self, doc, tag, ns_key=None, parent=None, check_validity=False, strict=DEFAULT_STRICT, exclude=()):
        if parent is None:
            parent = doc.getroot()
        if ns_key is None:
            node = create_new_node(doc, tag, parent=parent)
        else:
            node = create_new_node(doc, '{}:{}'.format(ns_key, tag), parent=parent)

        if 'RemapLUT' in self._child_xml_ns_key:
            rtag = '{}:RemapLUT'.format(self._child_xml_ns_key['RemapLUT'])
        elif ns_key is not None:
            rtag = '{}:RemapLUT'.format(ns_key)
        else:
            rtag = 'RemapLUT'

        if self._remap_lut is not None:
            value = ' '.join('{0:d},{1:d},{2:d}'.format(*entry) for entry in self._remap_lut)
            entry = create_text_node(doc, rtag, value, parent=node)
            entry.attrib['size'] = str(self.size)
        return node

    def to_dict(self,  check_validity=False, strict=DEFAULT_STRICT, exclude=()):
        out = OrderedDict()
        if self.RemapLUT is not None:
            out['RemapLUT'] = self.RemapLUT.tolist()
        return out


class MonochromeDisplayRemapType(Serializable):
    """
    This remap works by taking the input space and using the LUT to map it to a log space (for 8-bit only).

    From the log space the C0 and Ch fields are applied to get to display-ready density space. The density
    should then be rendered by the TTC and monitor comp. This means that the default DRA should not apply
    anything besides the clip points. If a different contrast/brightness is applied it should be done through
    modification of the clip points via DRA.
    """

    _fields = ('RemapType', 'RemapLUT', 'RemapParameters')
    _required = ('RemapType', )
    _collections_tags = {'RemapParameters': {'array': False, 'child_tag': 'RemapParameter'}}
    # Descriptor
    RemapType = StringDescriptor(
        'RemapType', _required, strict=DEFAULT_STRICT,
        docstring='')  # type: str
    RemapParameters = ParametersDescriptor(
        'RemapParameters', _collections_tags, _required, strict=DEFAULT_STRICT,
        docstring='Textual remap parameter. Filled based upon remap type (for informational purposes only).  '
                  'For example, if the data is linlog encoded a RemapParameter could be used to describe any '
                  'amplitude scaling that was performed prior to linlog encoding '
                  'the data.')  # type: Union[None, ParametersCollection]

    def __init__(self, RemapType=None, RemapLUT=None, RemapParameters=None, **kwargs):
        """

        Parameters
        ----------
        RemapType : str
        RemapLUT : None|numpy.ndarray
        RemapParameters : None|ParametersCollection|dict
        kwargs
        """

        self._remap_lut = None
        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.RemapType = RemapType
        self.RemapLUT = RemapLUT
        self.RemapParameters = RemapParameters
        super(MonochromeDisplayRemapType, self).__init__(**kwargs)

    @property
    def RemapLUT(self):
        """
        numpy.ndarray: the one dimensional Lookup table for remap to log amplitude for display,
        where the dtype must be `uint8`. Used during the "Product Generation Option" portion of the SIPS
        display chain. Required for 8-bit data, and not to be used for 16-bit data.
        """

        return self._remap_lut

    @RemapLUT.setter
    def RemapLUT(self, value):
        if value is None:
            self._remap_lut = None
            return
        if isinstance(value, (tuple, list)):
            value = numpy.array(value, dtype=numpy.uint8)
        if not isinstance(value, numpy.ndarray) or value.dtype.name != 'uint8':
            raise ValueError(
                'RemapLUT for class MonochromeDisplayRemapType must be a numpy.ndarray of dtype uint8.')
        if value.ndim != 1:
            raise ValueError('RemapLUT for class MonochromeDisplayRemapType must be a one-dimensional array.')
        self._remap_lut = value

    @classmethod
    def from_node(cls, node, xml_ns, ns_key=None, kwargs=None):
        if kwargs is None:
            kwargs = {}

        lut_key = cls._child_xml_ns_key.get('RemapLUT', ns_key)
        lut_node = find_first_child(node, 'RemapLUT', xml_ns, lut_key)
        if lut_node is not None:
            dim1 = int(lut_node.attrib['size'])
            arr = numpy.zeros((dim1, ), dtype=numpy.uint8)
            entries = get_node_value(lut_node).split()
            i = 0
            for entry in entries:
                if len(entry) == 0:
                    continue
                arr[i] = int(entry)
                i += 1
            kwargs['RemapLUT'] = arr
        return super(MonochromeDisplayRemapType, cls).from_node(node, xml_ns, ns_key=ns_key, **kwargs)

    def to_node(self, doc, tag, ns_key=None, parent=None, check_validity=False, strict=DEFAULT_STRICT, exclude=()):
        node = super(MonochromeDisplayRemapType, self).to_node(
            doc, tag, ns_key=ns_key, parent=parent, check_validity=check_validity, strict=strict)
        if 'RemapLUT' in self._child_xml_ns_key:
            rtag = '{}:RemapLUT'.format(self._child_xml_ns_key['RemapLUT'])
        elif ns_key is not None:
            rtag = '{}:RemapLUT'.format(ns_key)
        else:
            rtag = 'RemapLUT'

        if self._remap_lut is not None:
            value = ' '.join('{0:d}'.format(entry) for entry in self._remap_lut)
            entry = create_text_node(doc, rtag, value, parent=node)
            entry.attrib['size'] = str(self._remap_lut.size)
        return node

    def to_dict(self, check_validity=False, strict=DEFAULT_STRICT, exclude=()):
        out = super(MonochromeDisplayRemapType, self).to_dict(
            check_validity=check_validity, strict=strict, exclude=exclude)
        if self.RemapLUT is not None:
            out['RemapLUT'] = self.RemapLUT.tolist()
        return out


class RemapChoiceType(Serializable):
    """
    The remap choice type.
    """

    _fields = ('ColorDisplayRemap', 'MonochromeDisplayRemap')
    _required = ()
    _choice = ({'required': False, 'collection': ('ColorDisplayRemap', 'MonochromeDisplayRemap')}, )
    # Descriptor
    ColorDisplayRemap = SerializableDescriptor(
        'ColorDisplayRemap', ColorDisplayRemapType, _required, strict=DEFAULT_STRICT,
        docstring='Information for proper color display of the '
                  'data.')  # type: Union[None, ColorDisplayRemapType]
    MonochromeDisplayRemap = SerializableDescriptor(
        'MonochromeDisplayRemap', MonochromeDisplayRemapType, _required, strict=DEFAULT_STRICT,
        docstring='Information for proper monochrome display of the '
                  'data.')  # type: Union[None, MonochromeDisplayRemapType]

    def __init__(self, ColorDisplayRemap=None, MonochromeDisplayRemap=None, **kwargs):
        """

        Parameters
        ----------
        ColorDisplayRemap : None|ColorDisplayRemapType|numpy.ndarray|list|type
        MonochromeDisplayRemap : None|MonochromeDisplayRemapType
        kwargs
        """

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.ColorDisplayRemap = ColorDisplayRemap
        self.MonochromeDisplayRemap = MonochromeDisplayRemap
        super(RemapChoiceType, self).__init__(**kwargs)


class MonitorCompensationAppliedType(Serializable):
    """

    """
    _fields = ('Gamma', 'XMin')
    _required = ('Gamma', 'XMin')
    _numeric_format = {key: FLOAT_FORMAT for key in _fields}
    # Descriptor
    Gamma = FloatDescriptor(
        'Gamma', _required, strict=DEFAULT_STRICT,
        docstring='Gamma value for monitor compensation pre-applied to the image.')  # type: float
    XMin = FloatDescriptor(
        'XMin', _required, strict=DEFAULT_STRICT,
        docstring='Xmin value for monitor compensation pre-applied to the image.')  # type: float

    def __init__(self, Gamma=None, XMin=None, **kwargs):
        """

        Parameters
        ----------
        Gamma : float
        XMin : float
        kwargs
        """

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.Gamma = Gamma
        self.XMin = XMin
        super(MonitorCompensationAppliedType, self).__init__(**kwargs)


class DRAHistogramOverridesType(Serializable):
    """
    Dynamic range adjustment override parameters.
    """
    _fields = ('ClipMin', 'ClipMax')
    _required = ('ClipMin', 'ClipMax')
    # Descriptor
    ClipMin = IntegerDescriptor(
        'ClipMin', _required, strict=DEFAULT_STRICT,
        docstring='Suggested override for the lower end-point of the display histogram in the '
                  'ELT DRA application. Referred to as Pmin in SIPS documentation.')  # type: int
    ClipMax = IntegerDescriptor(
        'ClipMax', _required, strict=DEFAULT_STRICT,
        docstring='Suggested override for the upper end-point of the display histogram in the '
                  'ELT DRA application. Referred to as Pmax in SIPS documentation.')  # type: int

    def __init__(self, ClipMin=None, ClipMax=None, **kwargs):
        """

        Parameters
        ----------
        ClipMin : int
        ClipMax : int
        kwargs
        """

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.ClipMin = ClipMin
        self.ClipMax = ClipMax
        super(DRAHistogramOverridesType, self).__init__(**kwargs)


class ProductDisplayType(Serializable):
    """

    """
    _fields = (
        'PixelType', 'RemapInformation', 'MagnificationMethod', 'DecimationMethod',
        'DRAHistogramOverrides', 'MonitorCompensationApplied', 'DisplayExtensions')
    _required = ('PixelType', )
    _collections_tags = {
        'DisplayExtensions': {'array': False, 'child_tag': 'DisplayExtension'}}
    # Descriptors
    PixelType = StringEnumDescriptor(
        'PixelType', ('MONO8I', 'MONO8LU', 'MONO16I', 'RGB8LU', 'RGB24I'),
        _required, strict=DEFAULT_STRICT,
        docstring='Enumeration of the pixel type. Definition in '
                  'Design and Exploitation document.')  # type: str
    RemapInformation = SerializableDescriptor(
        'RemapInformation', RemapChoiceType, _required, strict=DEFAULT_STRICT,
        docstring='Information regarding the encoding of the pixel data. '
                  'Used for 8-bit pixel types.')  # type: Union[None, RemapChoiceType]
    MagnificationMethod = StringEnumDescriptor(
        'MagnificationMethod', ('NEAREST_NEIGHBOR', 'BILINEAR', 'LAGRANGE'),
        _required, strict=DEFAULT_STRICT,
        docstring='Recommended ELT magnification method for this data.')  # type: Union[None, str]
    DecimationMethod = StringEnumDescriptor(
        'DecimationMethod', ('NEAREST_NEIGHBOR', 'BILINEAR', 'BRIGHTEST_PIXEL', 'LAGRANGE'),
        _required, strict=DEFAULT_STRICT,
        docstring='Recommended ELT decimation method for this data. Also used as default for '
                  'reduced resolution dataset generation (if applicable).')  # type: Union[None, str]
    DRAHistogramOverrides = SerializableDescriptor(
        'DRAHistogramOverrides', DRAHistogramOverridesType, _required, strict=DEFAULT_STRICT,
        docstring='Recommended ELT DRA overrides.')  # type: Union[None, DRAHistogramOverridesType]
    MonitorCompensationApplied = SerializableDescriptor(
        'MonitorCompensationApplied', MonitorCompensationAppliedType, _required, strict=DEFAULT_STRICT,
        docstring='Describes monitor compensation that may have been applied to the product '
                  'during processing.')  # type: Union[None, MonitorCompensationAppliedType]
    DisplayExtensions = ParametersDescriptor(
        'DisplayExtensions', _collections_tags, _required, strict=DEFAULT_STRICT,
        docstring='Optional extensible parameters used to support profile-specific needs related to '
                  'product display. Predefined filter types.')  # type: Union[None, ParametersCollection]

    def __init__(self, PixelType=None, RemapInformation=None, MagnificationMethod=None, DecimationMethod=None,
                 DRAHistogramOverrides=None, MonitorCompensationApplied=None, DisplayExtensions=None, **kwargs):
        """

        Parameters
        ----------
        PixelType : PixelTypeType
        RemapInformation : None|RemapChoiceType
        MagnificationMethod : None|str
        DecimationMethod : None|str
        DRAHistogramOverrides : None|DRAHistogramOverridesType
        MonitorCompensationApplied : None|MonitorCompensationAppliedType
        DisplayExtensions : None|ParametersCollection|dict
        kwargs
        """

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.PixelType = PixelType
        self.RemapInformation = RemapInformation
        self.MagnificationMethod = MagnificationMethod
        self.DecimationMethod = DecimationMethod
        self.DRAHistogramOverrides = DRAHistogramOverrides
        self.MonitorCompensationApplied = MonitorCompensationApplied
        self.DisplayExtensions = DisplayExtensions
        super(ProductDisplayType, self).__init__(**kwargs)

    def get_pixel_size(self) -> int:
        """
        Gets the raw size per pixel, in bytes.

        Returns
        -------
        int
        """

        if self.PixelType == 'MONO8I':
            return 1
        elif self.PixelType == 'MONO8LU':
            return 1
        elif self.PixelType == 'MONO16I':
            return 2
        elif self.PixelType == 'RGB8LU':
            return 1
        elif self.PixelType == 'RGB24I':
            return 3
        else:
            raise ValueError('Got unhandled pixel type `{}`'.format(self.PixelType))



"""
The CollectionIDType definition.
"""

__classification__ = "UNCLASSIFIED"
__author__ = "Thomas McCullough"

from .base import DEFAULT_STRICT
from sarpy.io.complex.sicd_elements.CollectionInfo import CollectionInfoType
from sarpy.io.xml.descriptors import StringDescriptor


class CollectionIDType(CollectionInfoType):
    """
    The CollectionID type definition.
    """

    _fields = (
        'CollectorName', 'IlluminatorName', 'CoreName', 'CollectType',
        'RadarMode', 'Classification', 'ReleaseInfo', 'CountryCodes', 'Parameters')
    _required = ('CollectorName', 'CoreName', 'CollectType', 'RadarMode', 'Classification', 'ReleaseInfo')
    # descriptors
    ReleaseInfo = StringDescriptor(
        'ReleaseInfo', _required, strict=DEFAULT_STRICT, default_value='UNRESTRICTED',
        docstring='The product release information.')  # type: str

    def __init__(self, CollectorName=None, IlluminatorName=None, CoreName=None, CollectType=None,
                 RadarMode=None, Classification="UNCLASSIFIED", ReleaseInfo='UNRESTRICTED',
                 CountryCodes=None, Parameters=None, **kwargs):
        """

        Parameters
        ----------
        CollectorName : str
        IlluminatorName : str
        CoreName : str
        CollectType : str
        RadarMode : RadarModeType
        Classification : str
        ReleaseInfo : str
        CountryCodes : list|str
        Parameters : ParametersCollection|dict
        kwargs
        """

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.ReleaseInfo = ReleaseInfo
        super(CollectionIDType, self).__init__(
            CollectorName=CollectorName, IlluminatorName=IlluminatorName, CoreName=CoreName,
            CollectType=CollectType, RadarMode=RadarMode, Classification=Classification,
            CountryCodes=CountryCodes, Parameters=Parameters, **kwargs)



"""
The ExploitationFeaturesType definition for SIDD 1.0
"""

__classification__ = "UNCLASSIFIED"
__author__ = "Thomas McCullough"

import logging
import datetime
from typing import Union, List

import numpy

from sarpy.io.product.sidd2_elements.base import DEFAULT_STRICT, FLOAT_FORMAT

from sarpy.io.xml.base import Serializable, ParametersCollection
from sarpy.io.xml.descriptors import SerializableDescriptor, SerializableListDescriptor, \
    FloatDescriptor, FloatModularDescriptor, StringDescriptor, StringEnumDescriptor, \
    DateTimeDescriptor, ParametersDescriptor, BooleanDescriptor

from sarpy.io.product.sidd2_elements.blocks import RowColDoubleType, RangeAzimuthType, RadarModeType
# noinspection PyProtectedMember
from sarpy.io.product.sidd2_elements.ExploitationFeatures import InputROIType, \
    ExploitationFeaturesCollectionPhenomenologyType, \
    ExploitationCalculator, get_collection_resolution
from sarpy.io.complex.sicd_elements.SICD import SICDType

POLARIZATION_VALUES = ('V', 'H', 'RHC', 'LHC', 'OTHER')

logger = logging.getLogger(__name__)

_sicd_type_text = 'Requires instance of SICDType,\n\tgot type `{}`'


def _extract_sicd_tx_rcv_pol(str_in):
    """
    Extract the tx and rcv components from the sicd style tx/rcv polarization string.

    Parameters
    ----------
    str_in : str

    Returns
    -------
    str, str
    """

    if str_in is None:
        return 'OTHER', 'OTHER'

    if not isinstance(str_in, str):
        raise TypeError('requires a string type input.')

    if str_in in ['OTHER', 'UNKNOWN']:
        return 'OTHER', 'OTHER'

    def _decode_single_pol(pol):
        if pol in POLARIZATION_VALUES:
            return pol
        return 'OTHER'

    tx, rcv = str_in.split(':')
    tx = _decode_single_pol(tx)
    rcv = _decode_single_pol(rcv)

    return tx, rcv


class TxRcvPolarizationType(Serializable):
    """
    The transmit/receive polarization information.
    """

    _fields = ('TxPolarization', 'RcvPolarization', 'RcvPolarizationOffset', 'Processed')
    _required = ('TxPolarization', 'RcvPolarization')
    _numeric_format = {'RcvPolarizationOffset': FLOAT_FORMAT}
    # Descriptor
    TxPolarization = StringEnumDescriptor(
        'TxPolarization', POLARIZATION_VALUES, _required, strict=DEFAULT_STRICT,
        docstring='Transmit polarization type.')  # type: str
    RcvPolarization = StringEnumDescriptor(
        'RcvPolarization', POLARIZATION_VALUES, _required, strict=DEFAULT_STRICT,
        docstring='Receive polarization type.')  # type: str
    RcvPolarizationOffset = FloatModularDescriptor(
        'RcvPolarizationOffset', 180.0, _required, strict=DEFAULT_STRICT,
        docstring='Angle offset for the receive polarization defined at aperture '
                  'center.')  # type: Union[None, float]
    Processed = BooleanDescriptor(
        'Processed', _required, strict=DEFAULT_STRICT,
        docstring='')  # type: Union[None, bool]

    def __init__(self, TxPolarization=None, RcvPolarization=None, RcvPolarizationOffset=None,
                 Processed=None, **kwargs):
        """

        Parameters
        ----------
        TxPolarization : str
        RcvPolarization : str
        RcvPolarizationOffset : None|float
        Processed : None|bool
        kwargs
        """

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.TxPolarization = TxPolarization
        self.RcvPolarization = RcvPolarization
        self.RcvPolarizationOffset = RcvPolarizationOffset
        self.Processed = Processed
        super(TxRcvPolarizationType, self).__init__(**kwargs)

    @classmethod
    def from_sicd_value(cls, str_in):
        """
        Construct from the sicd style tx/rcv polarization string.

        Parameters
        ----------
        str_in : str

        Returns
        -------
        TxRcvPolarizationType
        """

        tx, rcv = _extract_sicd_tx_rcv_pol(str_in)
        return cls(TxPolarization=tx, RcvPolarization=rcv)


class ExploitationFeaturesCollectionInformationType(Serializable):
    """
    General collection information.
    """

    _fields = (
        'SensorName', 'RadarMode', 'CollectionDateTime', 'LocalDateTime', 'CollectionDuration',
        'Resolution', 'InputROI', 'Polarizations')
    _required = ('SensorName', 'RadarMode', 'CollectionDateTime', 'CollectionDuration')
    _collections_tags = {'Polarizations': {'array': False, 'child_tag': 'Polarization'}}
    _numeric_format = {'CollectionDuration': FLOAT_FORMAT}
    # Descriptor
    SensorName = StringDescriptor(
        'SensorName', _required, strict=DEFAULT_STRICT,
        docstring='The name of the sensor.')  # str
    RadarMode = SerializableDescriptor(
        'RadarMode', RadarModeType, _required, strict=DEFAULT_STRICT,
        docstring='Radar collection mode.')  # type: RadarModeType
    CollectionDateTime = DateTimeDescriptor(
        'CollectionDateTime', _required, strict=DEFAULT_STRICT,
        docstring='Collection date and time defined in Coordinated Universal Time (UTC). The seconds '
                  'should be followed by a Z to indicate UTC.')  # type: numpy.datetime64
    CollectionDuration = FloatDescriptor(
        'CollectionDuration', _required, strict=DEFAULT_STRICT,
        docstring='The duration of the collection (units = seconds).')  # type: float
    Resolution = SerializableDescriptor(
        'Resolution', RangeAzimuthType, _required, strict=DEFAULT_STRICT,
        docstring='Uniformly-weighted resolution (range and azimuth) processed in '
                  'the slant plane.')  # type: Union[None, RangeAzimuthType]
    InputROI = SerializableDescriptor(
        'InputROI', InputROIType, _required, strict=DEFAULT_STRICT,
        docstring='ROI representing portion of input data used to make '
                  'this product.')  # type: Union[None, InputROIType]
    Polarizations = SerializableListDescriptor(
        'Polarizations', TxRcvPolarizationType, _collections_tags, _required, strict=DEFAULT_STRICT,
        docstring='Transmit and receive polarization(s).')  # type: Union[None, List[TxRcvPolarizationType]]

    def __init__(self, SensorName=None, RadarMode=None, CollectionDateTime=None, LocalDateTime=None,
                 CollectionDuration=None, Resolution=None, Polarizations=None, **kwargs):
        """

        Parameters
        ----------
        SensorName : str
        RadarMode : RadarModeType
        CollectionDateTime : numpy.datetime64|datetime.datetime|datetime.date|str
        LocalDateTime : None|str|datetime.datetime
        CollectionDuration : float
        Resolution : None|RangeAzimuthType|numpy.ndarray|list|tuple
        Polarizations : None|List[TxTcvPolarizationType]
        kwargs
        """

        self._local_date_time = None
        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.SensorName = SensorName
        self.RadarMode = RadarMode
        self.CollectionDateTime = CollectionDateTime
        self.CollectionDuration = CollectionDuration
        self.LocalDateTime = LocalDateTime
        self.Resolution = Resolution
        self.Polarizations = Polarizations
        super(ExploitationFeaturesCollectionInformationType, self).__init__(**kwargs)

    @property
    def LocalDateTime(self):
        """None|str:  The local date/time string of the collection. *Optional.*"""
        return self._local_date_time

    @LocalDateTime.setter
    def LocalDateTime(self, value):  # type: (Union[None, str, datetime.datetime]) -> None
        if value is None:
            self._local_date_time = None
            return
        elif isinstance(value, datetime.datetime):
            value = value.isoformat('T')

        if isinstance(value, str):
            self._local_date_time = value
        else:
            logger.error(
                'Attribute LocalDateTime of class ExploitationFeaturesCollectionInformationType\n\t'
                'requires a datetime.datetime or string. Got unsupported type {}.\n\t'
                'Setting value to None.'.format(type(value)))
            self._local_date_time = None

    @classmethod
    def from_sicd(cls, sicd):
        """
        Construct from a sicd element.

        Parameters
        ----------
        sicd : SICDType

        Returns
        -------
        ExploitationFeaturesCollectionInformationType
        """

        if not isinstance(sicd, SICDType):
            raise TypeError(_sicd_type_text.format(type(sicd)))

        polarizations = [
            TxRcvPolarizationType.from_sicd_value(entry.TxRcvPolarization)
            for entry in sicd.RadarCollection.RcvChannels]

        return cls(SensorName=sicd.CollectionInfo.CollectorName,
                   RadarMode=RadarModeType(**sicd.CollectionInfo.RadarMode.to_dict()),
                   CollectionDateTime=sicd.Timeline.CollectStart,
                   CollectionDuration=sicd.Timeline.CollectDuration,
                   Resolution=get_collection_resolution(sicd),
                   Polarizations=polarizations)


class ExploitationFeaturesCollectionGeometryType(Serializable):
    """
    Key geometry parameters independent of product processing. All values
    computed at the center time of the full collection.
    """

    _fields = ('Azimuth', 'Slope', 'Squint', 'Graze', 'Tilt', 'Extensions')
    _required = ()
    _collections_tags = {'Extensions': {'array': False, 'child_tag': 'Extension'}}
    _numeric_format = {
        'Azimuth': FLOAT_FORMAT, 'Slope': FLOAT_FORMAT, 'Squint': FLOAT_FORMAT, 'Graze': FLOAT_FORMAT,
        'Tilt': FLOAT_FORMAT}
    # Descriptor
    Azimuth = FloatDescriptor(
        'Azimuth', _required, strict=DEFAULT_STRICT, bounds=(0.0, 360.0),
        docstring='Angle clockwise from north indicating the ETP line of sight vector.')  # type: float
    Slope = FloatDescriptor(
        'Slope', _required, strict=DEFAULT_STRICT, bounds=(0.0, 90.0),
        docstring='Angle between the ETP at scene center and the range vector perpendicular to '
                  'the direction of motion.')  # type: float
    Squint = FloatModularDescriptor(
        'Squint', 180.0, _required, strict=DEFAULT_STRICT,
        docstring='Angle from the ground track to platform velocity vector at nadir. '
                  'Left-look is positive, right-look is negative.')  # type: float
    Graze = FloatDescriptor(
        'Graze', _required, strict=DEFAULT_STRICT, bounds=(0.0, 90.0),
        docstring='Angle between the ETP and the line of sight vector.')  # type: float
    Tilt = FloatModularDescriptor(
        'Tilt', 180.0, _required, strict=DEFAULT_STRICT,
        docstring='Angle between the ETP and the cross range vector. '
                  'Also known as the twist angle.')  # type: float
    Extensions = ParametersDescriptor(
        'Extensions', _collections_tags, _required, strict=DEFAULT_STRICT,
        docstring='Exploitation feature extension related to geometry for a '
                  'single input image.')  # type: ParametersCollection

    def __init__(self, Azimuth=None, Slope=None, Squint=None, Graze=None, Tilt=None,
                 Extensions=None, **kwargs):
        """

        Parameters
        ----------
        Azimuth : None|float
        Slope : None|float
        Squint : None|float
        Graze : None|float
        Tilt : None|float
        Extensions : None|ParametersCollection|dict
        kwargs
        """

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.Azimuth = Azimuth
        self.Slope = Slope
        self.Squint = Squint
        self.Graze = Graze
        self.Tilt = Tilt
        self.Extensions = Extensions
        super(ExploitationFeaturesCollectionGeometryType, self).__init__(**kwargs)

    @classmethod
    def from_calculator(cls, calculator):
        """
        Create from an ExploitationCalculator object.

        Parameters
        ----------
        calculator : ExploitationCalculator

        Returns
        -------
        ExploitationFeaturesCollectionGeometryType
        """

        if not isinstance(calculator, ExploitationCalculator):
            raise TypeError(
                'Requires input which is an instance of ExploitationCalculator, got type {}'.format(type(calculator)))

        return cls(Azimuth=calculator.AzimuthAngle, Slope=calculator.SlopeAngle,
                   Squint=calculator.SquintAngle, Graze=calculator.GrazeAngle, Tilt=calculator.TiltAngle)


class CollectionType(Serializable):
    """
    Metadata regarding one of the input collections.
    """
    _fields = ('Information', 'Geometry', 'Phenomenology', 'identifier')
    _required = ('Information', 'identifier')
    _set_as_attribute = ('identifier', )
    # Descriptor
    Information = SerializableDescriptor(
        'Information', ExploitationFeaturesCollectionInformationType, _required, strict=DEFAULT_STRICT,
        docstring='General collection information.')  # type: ExploitationFeaturesCollectionInformationType
    Geometry = SerializableDescriptor(
        'Geometry', ExploitationFeaturesCollectionGeometryType, _required, strict=DEFAULT_STRICT,
        docstring='Key geometry parameters independent of product '
                  'processing.')  # type: Union[None, ExploitationFeaturesCollectionGeometryType]
    Phenomenology = SerializableDescriptor(
        'Phenomenology', ExploitationFeaturesCollectionPhenomenologyType, _required, strict=DEFAULT_STRICT,
        docstring='Phenomenology related to both the geometry and the final '
                  'product processing.')  # type: Union[None, ExploitationFeaturesCollectionPhenomenologyType]
    identifier = StringDescriptor(
        'identifier', _required, strict=DEFAULT_STRICT,
        docstring='The exploitation feature identifier.')  # type: str

    def __init__(self, Information=None, Geometry=None, Phenomenology=None, identifier=None, **kwargs):
        """

        Parameters
        ----------
        Information : ExploitationFeaturesCollectionInformationType
        Geometry : None|ExploitationFeaturesCollectionGeometryType
        Phenomenology : None|ExploitationFeaturesCollectionPhenomenologyType
        identifier : str
        kwargs
        """

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.Information = Information
        self.Geometry = Geometry
        self.Phenomenology = Phenomenology
        self.identifier = identifier
        super(CollectionType, self).__init__(**kwargs)

    @classmethod
    def from_calculator(cls, calculator, sicd):
        """
        Create from an ExploitationCalculator object.

        Parameters
        ----------
        calculator : ExploitationCalculator
        sicd : SICDType

        Returns
        -------
        CollectionType
        """

        if not isinstance(calculator, ExploitationCalculator):
            raise TypeError(
                'Requires input which is an instance of ExploitationCalculator, got type {}'.format(type(calculator)))

        return cls(identifier=sicd.CollectionInfo.CoreName,
                   Information=ExploitationFeaturesCollectionInformationType.from_sicd(sicd),
                   Geometry=ExploitationFeaturesCollectionGeometryType.from_calculator(calculator),
                   Phenomenology=ExploitationFeaturesCollectionPhenomenologyType.from_calculator(calculator))


class ExploitationFeaturesProductType(Serializable):
    """
    Metadata regarding the product.
    """

    _fields = ('Resolution', 'North', 'Extensions')
    _required = ('Resolution', )
    _collections_tags = {
        'Extensions': {'array': False, 'child_tag': 'Extension'}}
    _numeric_format = {'North': FLOAT_FORMAT}
    # Descriptor
    Resolution = SerializableDescriptor(
        'Resolution', RowColDoubleType, _required, strict=DEFAULT_STRICT,
        docstring='Uniformly-weighted resolution projected into the Earth Tangent '
                  'Plane (ETP).')  # type: RowColDoubleType
    North = FloatModularDescriptor(
        'North', 180.0, _required, strict=DEFAULT_STRICT,
        docstring='Counter-clockwise angle from increasing row direction to north at the center '
                  'of the image.')  # type: float
    Extensions = ParametersDescriptor(
        'Extensions', _collections_tags, _required, strict=DEFAULT_STRICT,
        docstring='Exploitation feature extension related to geometry for a '
                  'single input image.')  # type: ParametersCollection

    def __init__(self, Resolution=None, North=None, Extensions=None, **kwargs):
        """

        Parameters
        ----------
        Resolution : RowColDoubleType|numpy.ndarray|list|tuple
        North : None|float
        Extensions : None|ParametersCollection|dict
        kwargs
        """

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.Resolution = Resolution
        self.North = North
        self.Extensions = Extensions
        super(ExploitationFeaturesProductType, self).__init__(**kwargs)

    @classmethod
    def from_sicd(cls, sicd):
        """

        Parameters
        ----------
        sicd : SICDType

        Returns
        -------
        ExploitationFeaturesProductType
        """

        if not isinstance(sicd, SICDType):
            raise TypeError(_sicd_type_text.format(type(sicd)))

        row_ground, col_ground = sicd.get_ground_resolution()
        return cls(Resolution=(row_ground, col_ground))

    @classmethod
    def from_calculator(cls, calculator, sicd):
        """
        Construct from a sicd element and calculator.

        Parameters
        ----------
        calculator : ExploitationCalculator
        sicd : SICDType

        Returns
        -------
        ExploitationFeaturesProductType
        """

        if not isinstance(sicd, SICDType):
            raise TypeError(_sicd_type_text.format(type(sicd)))

        rho_rg_slant, rho_az_slant = get_collection_resolution(sicd)
        rho_row_etp, rho_col_etp = calculator.get_ground_plane_resolution(rho_rg_slant, rho_az_slant)

        return cls(Resolution=(rho_row_etp, rho_col_etp),
                   North=calculator.North)


class ExploitationFeaturesType(Serializable):
    """
    Computed metadata regarding the collect.
    """
    _fields = ('Collections', 'Product')
    _required = ('Collections', 'Product')
    _collections_tags = {
        'Collections': {'array': False, 'child_tag': 'Collection'}}
    # Descriptor
    Collections = SerializableListDescriptor(
        'Collections', CollectionType, _collections_tags, _required, strict=DEFAULT_STRICT,
        docstring='')  # type: List[CollectionType]
    Product = SerializableDescriptor(
        'Product', ExploitationFeaturesProductType, _required, strict=DEFAULT_STRICT,
        docstring='')  # type: ExploitationFeaturesProductType

    def __init__(self, Collections=None, Product=None, **kwargs):
        """

        Parameters
        ----------
        Collections : List[CollectionType]
        Product : ExploitationFeaturesProductType
        kwargs
        """

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.Collections = Collections
        self.Product = Product
        super(ExploitationFeaturesType, self).__init__(**kwargs)

    @classmethod
    def from_sicd(cls, sicd, row_vector, col_vector):
        """
        Construct from a sicd element.

        Parameters
        ----------
        sicd : SICDType|List[SICDType]
        row_vector : numpy.ndarray
        col_vector : numpy.ndarray

        Returns
        -------
        ExploitationFeaturesType
        """

        if isinstance(sicd, (list, tuple)):
            return cls(
                Collections=[
                    CollectionType.from_calculator(
                        ExploitationCalculator.from_sicd(entry, row_vector, col_vector), entry) for entry in sicd],
                Product=ExploitationFeaturesProductType.from_calculator(
                    ExploitationCalculator.from_sicd(sicd[0], row_vector, col_vector),
                    sicd[0]))

        if not isinstance(sicd, SICDType):
            raise TypeError(_sicd_type_text.format(type(sicd)))
        calculator = ExploitationCalculator.from_sicd(sicd, row_vector, col_vector)
        return cls(
            Collections=[
                CollectionType.from_calculator(calculator, sicd), ],
            Product=ExploitationFeaturesProductType.from_calculator(calculator, sicd))
